@@ -9,11 +9,13 @@ module GHC.Builtin.Names.TH where
 import GHC.Prelude ()
 
 import GHC.Builtin.Names( mk_known_key_name )
-import GHC.Unit
+import GHC.Unit.Types
+import GHC.Unit.Module.Name
 import GHC.Types.Name( Name )
 import GHC.Types.Name.Occurrence( tcName, clsName, dataName, varName )
 import GHC.Types.Name.Reader( RdrName, nameRdrName )
 import GHC.Types.Unique
+import GHC.Builtin.Uniques
 import GHC.Data.FastString
 
 -- To add a name, do three things
@@ -52,11 +54,11 @@ templateHaskellNames = [
     -- Exp
     varEName, conEName, litEName, appEName, appTypeEName, infixEName,
     infixAppName, sectionLName, sectionRName, lamEName, lamCaseEName,
-    tupEName, unboxedTupEName, unboxedSumEName,
+    lamCasesEName, tupEName, unboxedTupEName, unboxedSumEName,
     condEName, multiIfEName, letEName, caseEName, doEName, mdoEName, compEName,
     fromEName, fromThenEName, fromToEName, fromThenToEName,
     listEName, sigEName, recConEName, recUpdEName, staticEName, unboundVarEName,
-    labelEName, implicitParamVarEName,
+    labelEName, implicitParamVarEName, getFieldEName, projectionEName,
     -- FieldExp
     fieldExpName,
     -- Body
@@ -69,8 +71,8 @@ templateHaskellNames = [
     funDName, valDName, dataDName, newtypeDName, tySynDName,
     classDName, instanceWithOverlapDName,
     standaloneDerivWithStrategyDName, sigDName, kiSigDName, forImpDName,
-    pragInlDName, pragSpecDName, pragSpecInlDName, pragSpecInstDName,
-    pragRuleDName, pragCompleteDName, pragAnnDName, defaultSigDName,
+    pragInlDName, pragOpaqueDName, pragSpecDName, pragSpecInlDName, pragSpecInstDName,
+    pragRuleDName, pragCompleteDName, pragAnnDName, defaultSigDName, defaultDName,
     dataFamilyDName, openTypeFamilyDName, closedTypeFamilyDName,
     dataInstDName, newtypeInstDName, tySynInstDName,
     infixLDName, infixRDName, infixNDName,
@@ -102,7 +104,7 @@ templateHaskellNames = [
     promotedTName, promotedTupleTName, promotedNilTName, promotedConsTName,
     wildCardTName, implicitParamTName,
     -- TyLit
-    numTyLitName, strTyLitName,
+    numTyLitName, strTyLitName, charTyLitName,
     -- TyVarBndr
     plainTVName, kindedTVName,
     plainInvisTVName, kindedInvisTVName,
@@ -283,10 +285,10 @@ clauseName = libFun (fsLit "clause") clauseIdKey
 
 -- data Exp = ...
 varEName, conEName, litEName, appEName, appTypeEName, infixEName, infixAppName,
-    sectionLName, sectionRName, lamEName, lamCaseEName, tupEName,
+    sectionLName, sectionRName, lamEName, lamCaseEName, lamCasesEName, tupEName,
     unboxedTupEName, unboxedSumEName, condEName, multiIfEName, letEName,
     caseEName, doEName, mdoEName, compEName, staticEName, unboundVarEName,
-    labelEName, implicitParamVarEName :: Name
+    labelEName, implicitParamVarEName, getFieldEName, projectionEName :: Name
 varEName              = libFun (fsLit "varE")              varEIdKey
 conEName              = libFun (fsLit "conE")              conEIdKey
 litEName              = libFun (fsLit "litE")              litEIdKey
@@ -298,6 +300,7 @@ sectionLName          = libFun (fsLit "sectionL")          sectionLIdKey
 sectionRName          = libFun (fsLit "sectionR")          sectionRIdKey
 lamEName              = libFun (fsLit "lamE")              lamEIdKey
 lamCaseEName          = libFun (fsLit "lamCaseE")          lamCaseEIdKey
+lamCasesEName         = libFun (fsLit "lamCasesE")         lamCasesEIdKey
 tupEName              = libFun (fsLit "tupE")              tupEIdKey
 unboxedTupEName       = libFun (fsLit "unboxedTupE")       unboxedTupEIdKey
 unboxedSumEName       = libFun (fsLit "unboxedSumE")       unboxedSumEIdKey
@@ -324,6 +327,8 @@ staticEName           = libFun (fsLit "staticE")           staticEIdKey
 unboundVarEName       = libFun (fsLit "unboundVarE")       unboundVarEIdKey
 labelEName            = libFun (fsLit "labelE")            labelEIdKey
 implicitParamVarEName = libFun (fsLit "implicitParamVarE") implicitParamVarEIdKey
+getFieldEName         = libFun (fsLit "getFieldE")         getFieldEIdKey
+projectionEName       = libFun (fsLit "projectionE")       projectionEIdKey
 
 -- type FieldExp = ...
 fieldExpName :: Name
@@ -351,11 +356,11 @@ recSName    = libFun (fsLit "recS")    recSIdKey
 funDName, valDName, dataDName, newtypeDName, tySynDName, classDName,
     instanceWithOverlapDName, sigDName, kiSigDName, forImpDName, pragInlDName,
     pragSpecDName, pragSpecInlDName, pragSpecInstDName, pragRuleDName,
-    pragAnnDName, standaloneDerivWithStrategyDName, defaultSigDName,
+    pragAnnDName, standaloneDerivWithStrategyDName, defaultSigDName, defaultDName,
     dataInstDName, newtypeInstDName, tySynInstDName, dataFamilyDName,
     openTypeFamilyDName, closedTypeFamilyDName, infixLDName, infixRDName,
     infixNDName, roleAnnotDName, patSynDName, patSynSigDName,
-    pragCompleteDName, implicitParamBindDName :: Name
+    pragCompleteDName, implicitParamBindDName, pragOpaqueDName :: Name
 funDName                         = libFun (fsLit "funD")                         funDIdKey
 valDName                         = libFun (fsLit "valD")                         valDIdKey
 dataDName                        = libFun (fsLit "dataD")                        dataDIdKey
@@ -366,9 +371,11 @@ instanceWithOverlapDName         = libFun (fsLit "instanceWithOverlapD")
 standaloneDerivWithStrategyDName = libFun (fsLit "standaloneDerivWithStrategyD") standaloneDerivWithStrategyDIdKey
 sigDName                         = libFun (fsLit "sigD")                         sigDIdKey
 kiSigDName                       = libFun (fsLit "kiSigD")                       kiSigDIdKey
+defaultDName                     = libFun (fsLit "defaultD")                     defaultDIdKey
 defaultSigDName                  = libFun (fsLit "defaultSigD")                  defaultSigDIdKey
 forImpDName                      = libFun (fsLit "forImpD")                      forImpDIdKey
 pragInlDName                     = libFun (fsLit "pragInlD")                     pragInlDIdKey
+pragOpaqueDName                  = libFun (fsLit "pragOpaqueD")                  pragOpaqueDIdKey
 pragSpecDName                    = libFun (fsLit "pragSpecD")                    pragSpecDIdKey
 pragSpecInlDName                 = libFun (fsLit "pragSpecInlD")                 pragSpecInlDIdKey
 pragSpecInstDName                = libFun (fsLit "pragSpecInstD")                pragSpecInstDIdKey
@@ -468,9 +475,10 @@ infixTName          = libFun (fsLit "infixT")         infixTIdKey
 implicitParamTName  = libFun (fsLit "implicitParamT") implicitParamTIdKey
 
 -- data TyLit = ...
-numTyLitName, strTyLitName :: Name
+numTyLitName, strTyLitName, charTyLitName :: Name
 numTyLitName = libFun (fsLit "numTyLit") numTyLitIdKey
 strTyLitName = libFun (fsLit "strTyLit") strTyLitIdKey
+charTyLitName = libFun (fsLit "charTyLit") charTyLitIdKey
 
 -- data TyVarBndr = ...
 plainTVName, kindedTVName :: Name
@@ -701,14 +709,14 @@ inlinableDataConKey = mkPreludeDataConUnique 202
 
 -- data RuleMatch = ...
 conLikeDataConKey, funLikeDataConKey :: Unique
-conLikeDataConKey = mkPreludeDataConUnique 203
-funLikeDataConKey = mkPreludeDataConUnique 204
+conLikeDataConKey = mkPreludeDataConUnique 204
+funLikeDataConKey = mkPreludeDataConUnique 205
 
 -- data Phases = ...
 allPhasesDataConKey, fromPhaseDataConKey, beforePhaseDataConKey :: Unique
-allPhasesDataConKey   = mkPreludeDataConUnique 205
-fromPhaseDataConKey   = mkPreludeDataConUnique 206
-beforePhaseDataConKey = mkPreludeDataConUnique 207
+allPhasesDataConKey   = mkPreludeDataConUnique 206
+fromPhaseDataConKey   = mkPreludeDataConUnique 207
+beforePhaseDataConKey = mkPreludeDataConUnique 208
 
 -- data Overlap = ..
 overlappableDataConKey,
@@ -805,11 +813,12 @@ clauseIdKey         = mkPreludeMiscIdUnique 262
 -- data Exp = ...
 varEIdKey, conEIdKey, litEIdKey, appEIdKey, appTypeEIdKey, infixEIdKey,
     infixAppIdKey, sectionLIdKey, sectionRIdKey, lamEIdKey, lamCaseEIdKey,
-    tupEIdKey, unboxedTupEIdKey, unboxedSumEIdKey, condEIdKey, multiIfEIdKey,
-    letEIdKey, caseEIdKey, doEIdKey, compEIdKey,
+    lamCasesEIdKey, tupEIdKey, unboxedTupEIdKey, unboxedSumEIdKey, condEIdKey,
+    multiIfEIdKey, letEIdKey, caseEIdKey, doEIdKey, compEIdKey,
     fromEIdKey, fromThenEIdKey, fromToEIdKey, fromThenToEIdKey,
     listEIdKey, sigEIdKey, recConEIdKey, recUpdEIdKey, staticEIdKey,
-    unboundVarEIdKey, labelEIdKey, implicitParamVarEIdKey, mdoEIdKey :: Unique
+    unboundVarEIdKey, labelEIdKey, implicitParamVarEIdKey, mdoEIdKey,
+    getFieldEIdKey, projectionEIdKey :: Unique
 varEIdKey              = mkPreludeMiscIdUnique 270
 conEIdKey              = mkPreludeMiscIdUnique 271
 litEIdKey              = mkPreludeMiscIdUnique 272
@@ -821,50 +830,53 @@ sectionLIdKey          = mkPreludeMiscIdUnique 277
 sectionRIdKey          = mkPreludeMiscIdUnique 278
 lamEIdKey              = mkPreludeMiscIdUnique 279
 lamCaseEIdKey          = mkPreludeMiscIdUnique 280
-tupEIdKey              = mkPreludeMiscIdUnique 281
-unboxedTupEIdKey       = mkPreludeMiscIdUnique 282
-unboxedSumEIdKey       = mkPreludeMiscIdUnique 283
-condEIdKey             = mkPreludeMiscIdUnique 284
-multiIfEIdKey          = mkPreludeMiscIdUnique 285
-letEIdKey              = mkPreludeMiscIdUnique 286
-caseEIdKey             = mkPreludeMiscIdUnique 287
-doEIdKey               = mkPreludeMiscIdUnique 288
-compEIdKey             = mkPreludeMiscIdUnique 289
-fromEIdKey             = mkPreludeMiscIdUnique 290
-fromThenEIdKey         = mkPreludeMiscIdUnique 291
-fromToEIdKey           = mkPreludeMiscIdUnique 292
-fromThenToEIdKey       = mkPreludeMiscIdUnique 293
-listEIdKey             = mkPreludeMiscIdUnique 294
-sigEIdKey              = mkPreludeMiscIdUnique 295
-recConEIdKey           = mkPreludeMiscIdUnique 296
-recUpdEIdKey           = mkPreludeMiscIdUnique 297
-staticEIdKey           = mkPreludeMiscIdUnique 298
-unboundVarEIdKey       = mkPreludeMiscIdUnique 299
-labelEIdKey            = mkPreludeMiscIdUnique 300
-implicitParamVarEIdKey = mkPreludeMiscIdUnique 301
-mdoEIdKey              = mkPreludeMiscIdUnique 302
+lamCasesEIdKey         = mkPreludeMiscIdUnique 281
+tupEIdKey              = mkPreludeMiscIdUnique 282
+unboxedTupEIdKey       = mkPreludeMiscIdUnique 283
+unboxedSumEIdKey       = mkPreludeMiscIdUnique 284
+condEIdKey             = mkPreludeMiscIdUnique 285
+multiIfEIdKey          = mkPreludeMiscIdUnique 286
+letEIdKey              = mkPreludeMiscIdUnique 287
+caseEIdKey             = mkPreludeMiscIdUnique 288
+doEIdKey               = mkPreludeMiscIdUnique 289
+compEIdKey             = mkPreludeMiscIdUnique 290
+fromEIdKey             = mkPreludeMiscIdUnique 291
+fromThenEIdKey         = mkPreludeMiscIdUnique 292
+fromToEIdKey           = mkPreludeMiscIdUnique 293
+fromThenToEIdKey       = mkPreludeMiscIdUnique 294
+listEIdKey             = mkPreludeMiscIdUnique 295
+sigEIdKey              = mkPreludeMiscIdUnique 296
+recConEIdKey           = mkPreludeMiscIdUnique 297
+recUpdEIdKey           = mkPreludeMiscIdUnique 298
+staticEIdKey           = mkPreludeMiscIdUnique 299
+unboundVarEIdKey       = mkPreludeMiscIdUnique 300
+labelEIdKey            = mkPreludeMiscIdUnique 301
+implicitParamVarEIdKey = mkPreludeMiscIdUnique 302
+mdoEIdKey              = mkPreludeMiscIdUnique 303
+getFieldEIdKey         = mkPreludeMiscIdUnique 304
+projectionEIdKey       = mkPreludeMiscIdUnique 305
 
 -- type FieldExp = ...
 fieldExpIdKey :: Unique
-fieldExpIdKey       = mkPreludeMiscIdUnique 305
+fieldExpIdKey       = mkPreludeMiscIdUnique 306
 
 -- data Body = ...
 guardedBIdKey, normalBIdKey :: Unique
-guardedBIdKey     = mkPreludeMiscIdUnique 306
-normalBIdKey      = mkPreludeMiscIdUnique 307
+guardedBIdKey     = mkPreludeMiscIdUnique 307
+normalBIdKey      = mkPreludeMiscIdUnique 308
 
 -- data Guard = ...
 normalGEIdKey, patGEIdKey :: Unique
-normalGEIdKey     = mkPreludeMiscIdUnique 308
-patGEIdKey        = mkPreludeMiscIdUnique 309
+normalGEIdKey     = mkPreludeMiscIdUnique 309
+patGEIdKey        = mkPreludeMiscIdUnique 310
 
 -- data Stmt = ...
 bindSIdKey, letSIdKey, noBindSIdKey, parSIdKey, recSIdKey :: Unique
-bindSIdKey       = mkPreludeMiscIdUnique 310
-letSIdKey        = mkPreludeMiscIdUnique 311
-noBindSIdKey     = mkPreludeMiscIdUnique 312
-parSIdKey        = mkPreludeMiscIdUnique 313
-recSIdKey        = mkPreludeMiscIdUnique 314
+bindSIdKey       = mkPreludeMiscIdUnique 311
+letSIdKey        = mkPreludeMiscIdUnique 312
+noBindSIdKey     = mkPreludeMiscIdUnique 313
+parSIdKey        = mkPreludeMiscIdUnique 314
+recSIdKey        = mkPreludeMiscIdUnique 315
 
 -- data Dec = ...
 funDIdKey, valDIdKey, dataDIdKey, newtypeDIdKey, tySynDIdKey, classDIdKey,
@@ -875,7 +887,7 @@ funDIdKey, valDIdKey, dataDIdKey, newtypeDIdKey, tySynDIdKey, classDIdKey,
     newtypeInstDIdKey, tySynInstDIdKey, standaloneDerivWithStrategyDIdKey,
     infixLDIdKey, infixRDIdKey, infixNDIdKey, roleAnnotDIdKey, patSynDIdKey,
     patSynSigDIdKey, pragCompleteDIdKey, implicitParamBindDIdKey,
-    kiSigDIdKey :: Unique
+    kiSigDIdKey, defaultDIdKey, pragOpaqueDIdKey :: Unique
 funDIdKey                         = mkPreludeMiscIdUnique 320
 valDIdKey                         = mkPreludeMiscIdUnique 321
 dataDIdKey                        = mkPreludeMiscIdUnique 322
@@ -909,6 +921,8 @@ patSynSigDIdKey                   = mkPreludeMiscIdUnique 349
 pragCompleteDIdKey                = mkPreludeMiscIdUnique 350
 implicitParamBindDIdKey           = mkPreludeMiscIdUnique 351
 kiSigDIdKey                       = mkPreludeMiscIdUnique 352
+defaultDIdKey                     = mkPreludeMiscIdUnique 353
+pragOpaqueDIdKey                   = mkPreludeMiscIdUnique 354
 
 -- type Cxt = ...
 cxtIdKey :: Unique
@@ -989,14 +1003,15 @@ implicitParamTIdKey = mkPreludeMiscIdUnique 409
 infixTIdKey         = mkPreludeMiscIdUnique 410
 
 -- data TyLit = ...
-numTyLitIdKey, strTyLitIdKey :: Unique
-numTyLitIdKey = mkPreludeMiscIdUnique 411
-strTyLitIdKey = mkPreludeMiscIdUnique 412
+numTyLitIdKey, strTyLitIdKey, charTyLitIdKey :: Unique
+numTyLitIdKey  = mkPreludeMiscIdUnique 411
+strTyLitIdKey  = mkPreludeMiscIdUnique 412
+charTyLitIdKey = mkPreludeMiscIdUnique 413
 
 -- data TyVarBndr = ...
 plainTVIdKey, kindedTVIdKey :: Unique
-plainTVIdKey       = mkPreludeMiscIdUnique 413
-kindedTVIdKey      = mkPreludeMiscIdUnique 414
+plainTVIdKey       = mkPreludeMiscIdUnique 414
+kindedTVIdKey      = mkPreludeMiscIdUnique 415
 
 plainInvisTVIdKey, kindedInvisTVIdKey :: Unique
 plainInvisTVIdKey       = mkPreludeMiscIdUnique 482
@@ -1004,10 +1019,10 @@ kindedInvisTVIdKey      = mkPreludeMiscIdUnique 483
 
 -- data Role = ...
 nominalRIdKey, representationalRIdKey, phantomRIdKey, inferRIdKey :: Unique
-nominalRIdKey          = mkPreludeMiscIdUnique 415
-representationalRIdKey = mkPreludeMiscIdUnique 416
-phantomRIdKey          = mkPreludeMiscIdUnique 417
-inferRIdKey            = mkPreludeMiscIdUnique 418
+nominalRIdKey          = mkPreludeMiscIdUnique 416
+representationalRIdKey = mkPreludeMiscIdUnique 417
+phantomRIdKey          = mkPreludeMiscIdUnique 418
+inferRIdKey            = mkPreludeMiscIdUnique 419
 
 -- data Kind = ...
 starKIdKey, constraintKIdKey :: Unique
