@@ -1,5 +1,9 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FlexibleContexts #-}
+
 
 -- | Graph coloring register allocator.
 module GHC.CmmToAsm.Reg.Graph (
@@ -17,11 +21,13 @@ import GHC.CmmToAsm.Reg.Graph.TrivColorable
 import GHC.CmmToAsm.Instr
 import GHC.CmmToAsm.Reg.Target
 import GHC.CmmToAsm.Config
+import GHC.CmmToAsm.Types
 import GHC.Platform.Reg.Class
 import GHC.Platform.Reg
 
 import GHC.Data.Bag
 import GHC.Utils.Outputable
+import GHC.Utils.Panic
 import GHC.Platform
 import GHC.Types.Unique.FM
 import GHC.Types.Unique.Set
@@ -44,7 +50,7 @@ maxSpinCount    = 10
 
 -- | The top level of the graph coloring register allocator.
 regAlloc
-        :: (Outputable statics, Outputable instr, Instruction instr)
+        :: (OutputableP Platform statics, Instruction instr)
         => NCGConfig
         -> UniqFM RegClass (UniqSet RealReg)     -- ^ registers we can use for allocation
         -> UniqSet Int                  -- ^ set of available spill slots.
@@ -89,8 +95,7 @@ regAlloc config regsFree slotsFree slotsCount code cfg
 regAlloc_spin
         :: forall instr statics.
            (Instruction instr,
-            Outputable instr,
-            Outputable statics)
+            OutputableP Platform statics)
         => NCGConfig
         -> Int  -- ^ Number of solver iterations we've already performed.
         -> Color.Triv VirtualReg RegClass RealReg
@@ -388,7 +393,7 @@ graphAddCoalesce (r1, r2) graph
 
 -- | Patch registers in code using the reg -> reg mapping in this graph.
 patchRegsFromGraph
-        :: (Outputable statics, Outputable instr, Instruction instr)
+        :: (OutputableP Platform statics, Instruction instr)
         => Platform -> Color.Graph VirtualReg RegClass RealReg
         -> LiveCmmDecl statics instr -> LiveCmmDecl statics instr
 
@@ -413,7 +418,7 @@ patchRegsFromGraph platform graph code
                 = pprPanic "patchRegsFromGraph: register mapping failed."
                         (  text "There is no node in the graph for register "
                                 <> ppr reg
-                        $$ ppr code
+                        $$ pprLiveCmmDecl platform code
                         $$ Color.dotGraph
                                 (\_ -> text "white")
                                 (trivColorable platform
