@@ -41,11 +41,12 @@ import GHC.Prelude
 import GHC.Data.Graph.Base
 
 import GHC.Utils.Outputable
+import GHC.Utils.Panic
 import GHC.Types.Unique
 import GHC.Types.Unique.Set
 import GHC.Types.Unique.FM
 
-import Data.List        hiding (union)
+import Data.List (mapAccumL, sortBy)
 import Data.Maybe
 
 -- | Lookup a node from the graph.
