@@ -26,12 +26,11 @@ module GHC.StgToCmm.Layout (
         mkVirtConstrSizes,
         getHpRelOffset,
 
-        ArgRep(..), toArgRep, argRepSizeW -- re-exported from GHC.StgToCmm.ArgRep
+        ArgRep(..), toArgRep, argRepSizeW, -- re-exported from GHC.StgToCmm.ArgRep
+        getArgAmode, getNonVoidArgAmodes
   ) where
 
 
-#include "HsVersions.h"
-
 import GHC.Prelude hiding ((<*>))
 
 import GHC.StgToCmm.Closure
@@ -39,6 +38,7 @@ import GHC.StgToCmm.Env
 import GHC.StgToCmm.ArgRep -- notably: ( slowCallPattern )
 import GHC.StgToCmm.Ticky
 import GHC.StgToCmm.Monad
+import GHC.StgToCmm.Lit
 import GHC.StgToCmm.Utils
 
 import GHC.Cmm.Graph
@@ -52,15 +52,20 @@ import GHC.Stg.Syntax
 import GHC.Types.Id
 import GHC.Core.TyCon    ( PrimRep(..), primRepSizeB )
 import GHC.Types.Basic   ( RepArity )
-import GHC.Driver.Session
 import GHC.Platform
+import GHC.Platform.Profile
 import GHC.Unit
 
 import GHC.Utils.Misc
-import Data.List
+import Data.List (mapAccumL, partition)
 import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Utils.Panic.Plain
+import GHC.Utils.Constants (debugIsOn)
 import GHC.Data.FastString
 import Control.Monad
+import GHC.StgToCmm.Config (stgToCmmPlatform)
+import GHC.StgToCmm.Types
 
 ------------------------------------------------------------------------
 --                Call and return sequences
@@ -78,15 +83,15 @@ import Control.Monad
 --
 emitReturn :: [CmmExpr] -> FCode ReturnKind
 emitReturn results
-  = do { dflags    <- getDynFlags
+  = do { profile   <- getProfile
        ; platform  <- getPlatform
        ; sequel    <- getSequel
        ; updfr_off <- getUpdFrameOff
        ; case sequel of
            Return ->
              do { adjustHpBackwards
-                ; let e = CmmLoad (CmmStackSlot Old updfr_off) (gcWord platform)
-                ; emit (mkReturn dflags (entryCode platform e) results updfr_off)
+                ; let e = cmmLoadGCWord platform (CmmStackSlot Old updfr_off)
+                ; emit (mkReturn profile (entryCode platform e) results updfr_off)
                 }
            AssignTo regs adjust ->
              do { when adjust adjustHpBackwards
@@ -113,19 +118,19 @@ emitCallWithExtraStack
    :: (Convention, Convention) -> CmmExpr -> [CmmExpr]
    -> [CmmExpr] -> FCode ReturnKind
 emitCallWithExtraStack (callConv, retConv) fun args extra_stack
-  = do  { dflags <- getDynFlags
+  = do  { profile <- getProfile
         ; adjustHpBackwards
         ; sequel <- getSequel
         ; updfr_off <- getUpdFrameOff
         ; case sequel of
             Return -> do
-              emit $ mkJumpExtra dflags callConv fun args updfr_off extra_stack
+              emit $ mkJumpExtra profile callConv fun args updfr_off extra_stack
               return AssignedDirectly
             AssignTo res_regs _ -> do
               k <- newBlockId
               let area = Young k
-                  (off, _, copyin) = copyInOflow dflags retConv area res_regs []
-                  copyout = mkCallReturnsTo dflags fun callConv args k off updfr_off
+                  (off, _, copyin) = copyInOflow profile retConv area res_regs []
+                  copyout = mkCallReturnsTo profile fun callConv args k off updfr_off
                                    extra_stack
               tscope <- getTickScope
               emit (copyout <*> mkLabel k tscope <*> copyin)
@@ -190,8 +195,12 @@ directCall conv lbl arity stg_args
 slowCall :: CmmExpr -> [StgArg] -> FCode ReturnKind
 -- (slowCall fun args) applies fun to args, returning the results to Sequel
 slowCall fun stg_args
-  = do  dflags <- getDynFlags
-        platform <- getPlatform
+  = do  cfg <- getStgToCmmConfig
+        let profile   = stgToCmmProfile      cfg
+            platform  = stgToCmmPlatform     cfg
+            ctx       = stgToCmmContext      cfg
+            fast_pap  = stgToCmmFastPAPCalls cfg
+            align_sat = stgToCmmAlignCheck   cfg
         argsreps <- getArgRepsAmodes stg_args
         let (rts_fun, arity) = slowCallPattern (map fst argsreps)
 
@@ -199,17 +208,17 @@ slowCall fun stg_args
            r <- direct_call "slow_call" NativeNodeCall
                  (mkRtsApFastLabel rts_fun) arity ((P,Just fun):argsreps)
            emitComment $ mkFastString ("slow_call for " ++
-                                      showSDoc dflags (ppr fun) ++
+                                      renderWithContext ctx (pdoc platform fun) ++
                                       " with pat " ++ unpackFS rts_fun)
            return r
 
-        -- Note [avoid intermediate PAPs]
+        -- See Note [avoid intermediate PAPs]
         let n_args = length stg_args
-        if n_args > arity && optLevel dflags >= 2
+        if n_args > arity && fast_pap
            then do
              funv <- (CmmReg . CmmLocal) `fmap` assignTemp fun
              fun_iptr <- (CmmReg . CmmLocal) `fmap`
-                    assignTemp (closureInfoPtr dflags (cmmUntag dflags funv))
+               assignTemp (closureInfoPtr platform align_sat (cmmUntag platform funv))
 
              -- ToDo: we could do slightly better here by reusing the
              -- continuation from the slow call, which we have in r.
@@ -230,11 +239,11 @@ slowCall fun stg_args
              is_tagged_lbl <- newBlockId
              end_lbl <- newBlockId
 
-             let correct_arity = cmmEqWord platform (funInfoArity dflags fun_iptr)
+             let correct_arity = cmmEqWord platform (funInfoArity profile fun_iptr)
                                                     (mkIntExpr platform n_args)
 
              tscope <- getTickScope
-             emit (mkCbranch (cmmIsTagged dflags funv)
+             emit (mkCbranch (cmmIsTagged platform funv)
                              is_tagged_lbl slow_lbl (Just True)
                    <*> mkLabel is_tagged_lbl tscope
                    <*> mkCbranch correct_arity fast_lbl slow_lbl (Just True)
@@ -252,7 +261,7 @@ slowCall fun stg_args
 
 
 -- Note [avoid intermediate PAPs]
---
+-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 -- A slow call which needs multiple generic apply patterns will be
 -- almost guaranteed to create one or more intermediate PAPs when
 -- applied to a function that takes the correct number of arguments.
@@ -285,24 +294,24 @@ direct_call :: String
 direct_call caller call_conv lbl arity args
   | debugIsOn && args `lengthLessThan` real_arity  -- Too few args
   = do -- Caller should ensure that there enough args!
+       platform <- getPlatform
        pprPanic "direct_call" $
             text caller <+> ppr arity <+>
-            ppr lbl <+> ppr (length args) <+>
-            ppr (map snd args) <+> ppr (map fst args)
+            pdoc platform lbl <+> ppr (length args) <+>
+            pdoc platform (map snd args) <+> ppr (map fst args)
 
   | null rest_args  -- Precisely the right number of arguments
   = emitCall (call_conv, NativeReturn) target (nonVArgs args)
 
   | otherwise       -- Note [over-saturated calls]
-  = do dflags <- getDynFlags
+  = do do_scc_prof <- stgToCmmSCCProfiling <$> getStgToCmmConfig
        emitCallWithExtraStack (call_conv, NativeReturn)
                               target
                               (nonVArgs fast_args)
-                              (nonVArgs (stack_args dflags))
+                              (nonVArgs (slowArgs rest_args do_scc_prof))
   where
     target = CmmLit (CmmLabel lbl)
     (fast_args, rest_args) = splitAt real_arity args
-    stack_args dflags = slowArgs dflags rest_args
     real_arity = case call_conv of
                    NativeNodeCall -> arity+1
                    _              -> arity
@@ -314,12 +323,14 @@ direct_call caller call_conv lbl arity args
 -- using zeroCLit or even undefined would work, but would be ugly).
 --
 getArgRepsAmodes :: [StgArg] -> FCode [(ArgRep, Maybe CmmExpr)]
-getArgRepsAmodes = mapM getArgRepAmode
-  where getArgRepAmode arg
+getArgRepsAmodes args = do
+   platform <- profilePlatform <$> getProfile
+   mapM (getArgRepAmode platform) args
+  where getArgRepAmode platform arg
            | V <- rep  = return (V, Nothing)
            | otherwise = do expr <- getArgAmode (NonVoid arg)
                             return (rep, Just expr)
-           where rep = toArgRep (argPrimRep arg)
+           where rep = toArgRep platform (argPrimRep arg)
 
 nonVArgs :: [(ArgRep, Maybe CmmExpr)] -> [CmmExpr]
 nonVArgs [] = []
@@ -328,7 +339,7 @@ nonVArgs ((_,Just arg) : args) = arg : nonVArgs args
 
 {-
 Note [over-saturated calls]
-
+~~~~~~~~~~~~~~~~~~~~~~~~~~~
 The natural thing to do for an over-saturated call would be to call
 the function with the correct number of arguments, and then apply the
 remaining arguments to the value returned, e.g.
@@ -364,12 +375,11 @@ just more arguments that we are passing on the stack (cml_args).
 -- | 'slowArgs' takes a list of function arguments and prepares them for
 -- pushing on the stack for "extra" arguments to a function which requires
 -- fewer arguments than we currently have.
-slowArgs :: DynFlags -> [(ArgRep, Maybe CmmExpr)] -> [(ArgRep, Maybe CmmExpr)]
-slowArgs _ [] = []
-slowArgs dflags args -- careful: reps contains voids (V), but args does not
-  | sccProfilingEnabled dflags
-              = save_cccs ++ this_pat ++ slowArgs dflags rest_args
-  | otherwise =              this_pat ++ slowArgs dflags rest_args
+slowArgs :: [(ArgRep, Maybe CmmExpr)] -> DoSCCProfiling -> [(ArgRep, Maybe CmmExpr)]
+slowArgs []   _                    = mempty
+slowArgs args sccProfilingEnabled  -- careful: reps contains voids (V), but args does not
+  | sccProfilingEnabled = save_cccs ++ this_pat ++ slowArgs rest_args sccProfilingEnabled
+  | otherwise           =              this_pat ++ slowArgs rest_args sccProfilingEnabled
   where
     (arg_pat, n)            = slowCallPattern (map fst args)
     (call_args, rest_args)  = splitAt n args
@@ -411,7 +421,7 @@ data ClosureHeader
   | ThunkHeader
 
 mkVirtHeapOffsetsWithPadding
-  :: DynFlags
+  :: Profile
   -> ClosureHeader            -- What kind of header to account for
   -> [NonVoid (PrimRep, a)]   -- Things to make offsets for
   -> ( WordOff                -- Total number of words allocated
@@ -426,18 +436,18 @@ mkVirtHeapOffsetsWithPadding
 -- mkVirtHeapOffsetsWithPadding always returns boxed things with smaller offsets
 -- than the unboxed things
 
-mkVirtHeapOffsetsWithPadding dflags header things =
-    ASSERT(not (any (isVoidRep . fst . fromNonVoid) things))
+mkVirtHeapOffsetsWithPadding profile header things =
+    assert (not (any (isVoidRep . fst . fromNonVoid) things))
     ( tot_wds
     , bytesToWordsRoundUp platform bytes_of_ptrs
     , concat (ptrs_w_offsets ++ non_ptrs_w_offsets) ++ final_pad
     )
   where
-    platform = targetPlatform dflags
+    platform = profilePlatform profile
     hdr_words = case header of
       NoHeader -> 0
-      StdHeader -> fixedHdrSizeW dflags
-      ThunkHeader -> thunkHdrSize dflags
+      StdHeader -> fixedHdrSizeW profile
+      ThunkHeader -> thunkHdrSize profile
     hdr_bytes = wordsToBytes platform hdr_words
 
     (ptrs, non_ptrs) = partition (isGcPtrRep . fst . fromNonVoid) things
@@ -485,36 +495,36 @@ mkVirtHeapOffsetsWithPadding dflags header things =
 
 
 mkVirtHeapOffsets
-  :: DynFlags
+  :: Profile
   -> ClosureHeader            -- What kind of header to account for
   -> [NonVoid (PrimRep,a)]    -- Things to make offsets for
   -> (WordOff,                -- _Total_ number of words allocated
       WordOff,                -- Number of words allocated for *pointers*
       [(NonVoid a, ByteOff)])
-mkVirtHeapOffsets dflags header things =
+mkVirtHeapOffsets profile header things =
     ( tot_wds
     , ptr_wds
     , [ (field, offset) | (FieldOff field offset) <- things_offsets ]
     )
   where
    (tot_wds, ptr_wds, things_offsets) =
-       mkVirtHeapOffsetsWithPadding dflags header things
+       mkVirtHeapOffsetsWithPadding profile header things
 
 -- | Just like mkVirtHeapOffsets, but for constructors
 mkVirtConstrOffsets
-  :: DynFlags -> [NonVoid (PrimRep, a)]
+  :: Profile -> [NonVoid (PrimRep, a)]
   -> (WordOff, WordOff, [(NonVoid a, ByteOff)])
-mkVirtConstrOffsets dflags = mkVirtHeapOffsets dflags StdHeader
+mkVirtConstrOffsets profile = mkVirtHeapOffsets profile StdHeader
 
 -- | Just like mkVirtConstrOffsets, but used when we don't have the actual
 -- arguments. Useful when e.g. generating info tables; we just need to know
 -- sizes of pointer and non-pointer fields.
-mkVirtConstrSizes :: DynFlags -> [NonVoid PrimRep] -> (WordOff, WordOff)
-mkVirtConstrSizes dflags field_reps
+mkVirtConstrSizes :: Profile -> [NonVoid PrimRep] -> (WordOff, WordOff)
+mkVirtConstrSizes profile field_reps
   = (tot_wds, ptr_wds)
   where
     (tot_wds, ptr_wds, _) =
-       mkVirtConstrOffsets dflags
+       mkVirtConstrOffsets profile
          (map (\nv_rep -> NonVoid (fromNonVoid nv_rep, ())) field_reps)
 
 -------------------------------------------------------------------------
@@ -530,12 +540,12 @@ mkVirtConstrSizes dflags field_reps
 -------------------------------------------------------------------------
 
 -- bring in ARG_P, ARG_N, etc.
-#include "../includes/rts/storage/FunTypes.h"
+#include "FunTypes.h"
 
 mkArgDescr :: Platform -> [Id] -> ArgDescr
 mkArgDescr platform args
   = let arg_bits = argBits platform arg_reps
-        arg_reps = filter isNonV (map idArgRep args)
+        arg_reps = filter isNonV (map (idArgRep platform) args)
            -- Getting rid of voids eases matching of standard patterns
     in case stdPattern arg_reps of
          Just spec_id -> ArgSpec spec_id
@@ -581,6 +591,24 @@ stdPattern reps
 
         _ -> Nothing
 
+-------------------------------------------------------------------------
+--        Amodes for arguments
+-------------------------------------------------------------------------
+
+getArgAmode :: NonVoid StgArg -> FCode CmmExpr
+getArgAmode (NonVoid (StgVarArg var)) = idInfoToAmode <$> getCgIdInfo var
+getArgAmode (NonVoid (StgLitArg lit)) = cgLit lit
+
+getNonVoidArgAmodes :: [StgArg] -> FCode [CmmExpr]
+-- NB: Filters out void args,
+--     so the result list may be shorter than the argument list
+getNonVoidArgAmodes [] = return []
+getNonVoidArgAmodes (arg:args)
+  | isVoidRep (argPrimRep arg) = getNonVoidArgAmodes args
+  | otherwise = do { amode  <- getArgAmode (NonVoid arg)
+                   ; amodes <- getNonVoidArgAmodes args
+                   ; return ( amode : amodes ) }
+
 -------------------------------------------------------------------------
 --
 --        Generating the info table and code for a closure
@@ -601,28 +629,28 @@ emitClosureProcAndInfoTable :: Bool                    -- top-level?
                             -> ((Int, LocalReg, [LocalReg]) -> FCode ()) -- function body
                             -> FCode ()
 emitClosureProcAndInfoTable top_lvl bndr lf_info info_tbl args body
- = do   { dflags <- getDynFlags
+ = do   { profile <- getProfile
         ; platform <- getPlatform
         -- Bind the binder itself, but only if it's not a top-level
         -- binding. We need non-top let-bindings to refer to the
         -- top-level binding, which this binding would incorrectly shadow.
         ; node <- if top_lvl then return $ idToReg platform (NonVoid bndr)
                   else bindToReg (NonVoid bndr) lf_info
-        ; let node_points = nodeMustPointToIt dflags lf_info
+        ; let node_points = nodeMustPointToIt profile lf_info
         ; arg_regs <- bindArgsToRegs args
         ; let args' = if node_points then (node : arg_regs) else arg_regs
-              conv  = if nodeMustPointToIt dflags lf_info then NativeNodeCall
+              conv  = if nodeMustPointToIt profile lf_info then NativeNodeCall
                                                           else NativeDirectCall
-              (offset, _, _) = mkCallEntry dflags conv args' []
-        ; emitClosureAndInfoTable info_tbl conv args' $ body (offset, node, arg_regs)
+              (offset, _, _) = mkCallEntry profile conv args' []
+        ; emitClosureAndInfoTable (profilePlatform profile) info_tbl conv args' $ body (offset, node, arg_regs)
         }
 
 -- Data constructors need closures, but not with all the argument handling
 -- needed for functions. The shared part goes here.
-emitClosureAndInfoTable ::
-  CmmInfoTable -> Convention -> [LocalReg] -> FCode () -> FCode ()
-emitClosureAndInfoTable info_tbl conv args body
+emitClosureAndInfoTable
+   :: Platform -> CmmInfoTable -> Convention -> [LocalReg] -> FCode () -> FCode ()
+emitClosureAndInfoTable platform info_tbl conv args body
   = do { (_, blks) <- getCodeScoped body
-       ; let entry_lbl = toEntryLbl (cit_lbl info_tbl)
+       ; let entry_lbl = toEntryLbl platform (cit_lbl info_tbl)
        ; emitProcWithConvention conv (Just info_tbl) entry_lbl args blks
        }
