@@ -11,9 +11,13 @@ import Foreign.Ptr (ptrToIntPtr, intPtrToPtr)
 import GHC
 import GHC.Ptr (Ptr (..))
 import GHCi.Util
-import GHC.Driver.Types
+import GHC.Driver.Env
+import GHC.Driver.Ppr
 import GHC.Utils.Outputable
+import GHC.Unit.Module.ModDetails
+import GHC.Unit.Home.ModInfo
 import GHC.Platform (target32Bit)
+import GHC.Linker.Types
 import Prelude
 import System.Mem
 import System.Mem.Weak
@@ -34,9 +38,9 @@ data LeakModIndicators = LeakModIndicators
 -- | Grab weak references to some of the data structures representing
 -- the currently loaded modules.
 getLeakIndicators :: HscEnv -> IO LeakIndicators
-getLeakIndicators HscEnv{..} =
+getLeakIndicators hsc_env =
   fmap LeakIndicators $
-    forM (eltsUDFM hsc_HPT) $ \hmi@HomeModInfo{..} -> do
+    forM (eltsUDFM (hsc_HPT hsc_env)) $ \hmi@HomeModInfo{..} -> do
       leakMod <- mkWeakPtr hmi Nothing
       leakIface <- mkWeakPtr hm_iface Nothing
       leakDetails <- mkWeakPtr hm_details Nothing
@@ -55,7 +59,9 @@ checkLeakIndicators dflags (LeakIndicators leakmods)  = do
       Just hmi ->
         report ("HomeModInfo for " ++
           showSDoc dflags (ppr (mi_module (hm_iface hmi)))) (Just hmi)
-    deRefWeak leakIface >>= report "ModIface"
+    deRefWeak leakIface >>= \case
+      Nothing -> return ()
+      Just miface -> report ("ModIface:" ++ moduleNameString (moduleName (mi_module miface))) (Just miface)
     deRefWeak leakDetails >>= report "ModDetails"
     forM_ leakLinkable $ \l -> deRefWeak l >>= report "Linkable"
  where
