@@ -14,6 +14,7 @@ import qualified Prelude as IO (writeFile)
 
 import Distribution.Compat.Binary
 import Distribution.Simple.Utils (withTempDirectory)
+import Distribution.System (buildOS, OS (Windows))
 import Distribution.Verbosity (silent)
 
 import Distribution.Client.FileMonitor
@@ -22,6 +23,7 @@ import Distribution.Utils.Structured (structureHash, Structured)
 import GHC.Fingerprint (Fingerprint (..))
 
 import Test.Tasty
+import Test.Tasty.ExpectedFailure
 import Test.Tasty.HUnit
 
 
@@ -60,8 +62,9 @@ tests mtimeChange =
     , testCase "add non-match"       $ testGlobAddNonMatch mtimeChange
     , testCase "remove non-match"    $ testGlobRemoveNonMatch mtimeChange
 
-    , testCase "add non-match"       $ testGlobAddNonMatchSubdir mtimeChange
-    , testCase "remove non-match"    $ testGlobRemoveNonMatchSubdir mtimeChange
+    , knownBrokenInWindows "See issue #3126" $
+      testCase "add non-match subdir"    $ testGlobAddNonMatchSubdir mtimeChange
+    , testCase "remove non-match subdir" $ testGlobRemoveNonMatchSubdir mtimeChange
 
     , testCase "invariant sorted 1"  $ testInvariantMonitorStateGlobFiles
                                          mtimeChange
@@ -69,7 +72,8 @@ tests mtimeChange =
                                          mtimeChange
 
     , testCase "match dirs"          $ testGlobMatchDir mtimeChange
-    , testCase "match dirs only"     $ testGlobMatchDirOnly mtimeChange
+    , knownBrokenInWindows "See issue #3126" $
+      testCase "match dirs only"     $ testGlobMatchDirOnly mtimeChange
     , testCase "change file type"    $ testGlobChangeFileType mtimeChange
     , testCase "absolute paths"      $ testGlobAbsolutePath mtimeChange
     ]
@@ -80,6 +84,10 @@ tests mtimeChange =
   , testCase "value updated"         testValueUpdated
   ]
 
+  where knownBrokenInWindows msg =  case buildOS of
+          Windows -> expectFailBecause msg
+          _       -> id
+
 -- Check the file system behaves the way we expect it to
 
 -- we rely on file mtimes having a reasonable resolution
@@ -697,19 +705,19 @@ testGlobAbsolutePath mtimeChange =
     threadDelay mtimeChange
     removeFile root "dir/good-a"
     reason <- expectMonitorChanged root monitor ()
-    reason @?= MonitoredFileChanged (root' </> "dir/good-a")
+    reason @?= MonitoredFileChanged (root' </> "dir" </> "good-a")
     -- absolute glob, adding a file
     updateMonitor root monitor [monitorFileGlobStr (root' </> "dir/good-*")] () ()
     threadDelay mtimeChange
     touchFile root ("dir/good-a")
     reason2 <- expectMonitorChanged root monitor ()
-    reason2 @?= MonitoredFileChanged (root' </> "dir/good-a")
+    reason2 @?= MonitoredFileChanged (root' </> "dir" </> "good-a")
     -- absolute glob, changing a file
     updateMonitor root monitor [monitorFileGlobStr (root' </> "dir/good-*")] () ()
     threadDelay mtimeChange
     touchFileContent root "dir/good-b"
     reason3 <- expectMonitorChanged root monitor ()
-    reason3 @?= MonitoredFileChanged (root' </> "dir/good-b")
+    reason3 @?= MonitoredFileChanged (root' </> "dir" </> "good-b")
 
 
 ------------------
