@@ -1,6 +1,6 @@
 import Test.Cabal.Prelude
 main = setupAndCabalTest $ do
-    skipUnless =<< ghcVersionIs (>= mkVersion [8,1])
-    skipUnless =<< hasProfiledLibraries
+    skipUnlessGhcVersion ">= 8.1"
+    skipUnless "no profiling libs" =<< hasProfiledLibraries
     setup "configure" ["--enable-profiling"]
     setup "build" []
