@@ -4,7 +4,13 @@
 # default config
 #######################################################################
 
-HC=ghc-8.2.2
+# We use the default ghc in PATH as default
+# Use the ghc-x.y.z trigger several errors in windows:
+# * It triggers the max path length issue:
+#   See https://github.com/haskell/cabal/issues/6271#issuecomment-1065102255
+# * It triggers a `createProcess: does not exist` error in units tests
+#   See https://github.com/haskell/cabal/issues/8049
+HC=ghc
 CABAL=cabal
 CABALPLAN=cabal-plan
 JOBS=4
@@ -36,7 +42,7 @@ Usage: ./validate.sh [options]
 
 Available options:
   -j, --jobs JOBS                   cabal v2-build -j argument (default:  $JOBS)
-      --libonly                     Test onlt Cabal-the-library
+      --libonly                     Test only Cabal-the-library
       --cli                         Test both Cabal-the-library and cabal-install
       --(no-)run-lib-tests          Run library tests
       --(no-)run-cli-tests          Run client tests
@@ -52,7 +58,7 @@ Available options:
       --partial-hackage-tests       Run hackage-tests on parts of Hackage data
   -v, --verbose                     Verbose output
   -q, --quiet                       Less output
-  -s, --step STEP                   Run only specific step (can be specified mutliple times)
+  -s, --step STEP                   Run only specific step (can be specified multiple times)
       --list-steps                  List steps and build-targets and exit
       --help                        Print this message and exit
 EOF
@@ -259,7 +265,6 @@ fi
 
 if [ -z "$STEPS" ]; then
     STEPS="print-config print-tool-versions"
-    if ! $LIBONLY;  then STEPS="$STEPS make-cabal-install-dev"; fi
     STEPS="$STEPS build"
     if $DOCTEST;    then STEPS="$STEPS doctest";   fi
     if $LIBTESTS;   then STEPS="$STEPS lib-tests"; fi
@@ -272,8 +277,8 @@ if [ -z "$STEPS" ]; then
     STEPS="$STEPS time-summary"
 fi
 
-TARGETS="Cabal cabal-testsuite"
-if ! $LIBONLY;  then TARGETS="$TARGETS cabal-install"; fi
+TARGETS="Cabal cabal-testsuite Cabal-tests Cabal-QuickCheck Cabal-tree-diff Cabal-described"
+if ! $LIBONLY;  then TARGETS="$TARGETS cabal-install cabal-install-solver cabal-benchmarks"; fi
 if $BENCHMARKS; then TARGETS="$TARGETS solver-benchmarks"; fi
 
 if $LISTSTEPS; then
@@ -291,7 +296,9 @@ JOBS="-j$JOBS"
 # assume compiler is GHC
 RUNHASKELL=$(echo $HC | sed -E 's/ghc(-[0-9.]*)$/runghc\1/')
 
-if [ "$(uname)" = "Linux" ]; then
+if [ "$OSTYPE" = "msys" ]; then
+    ARCH="x86_64-windows"
+elif [ "$(uname)" = "Linux" ]; then
     ARCH="x86_64-linux"
 else
     ARCH="x86_64-osx"
@@ -318,7 +325,7 @@ print_header print-config
 
 cat <<EOF
 compiler:            $HC
-runhaskell           $RUNHASKELL
+runhaskell:          $RUNHASKELL
 cabal-install:       $CABAL
 cabal-plan:          $CABALPLAN
 jobs:                $JOBS
@@ -330,7 +337,7 @@ dependencies only:   $DEPSONLY
 doctest:             $DOCTEST
 benchmarks:          $BENCHMARKS
 verbose:             $VERBOSE
-extra complers:      $EXTRAHCS
+extra compilers:     $EXTRAHCS
 
 EOF
 }
@@ -359,11 +366,6 @@ step_time_summary() {
 # build
 #######################################################################
 
-step_make_cabal_install_dev() {
-print_header "make cabal-install-dev"
-timed cp cabal-install/cabal-install.cabal.dev cabal-install/cabal-install.cabal
-}
-
 step_build() {
 print_header "build"
 timed $CABALNEWBUILD $TARGETS --dry-run || exit 1
@@ -383,24 +385,29 @@ timed doctest -package-env=doctest-Cabal --fast Cabal/Distribution Cabal/Languag
 step_lib_tests() {
 print_header "Cabal: tests"
 
-CMD="$($CABALPLANLISTBIN Cabal:test:unit-tests) $TESTSUITEJOBS --hide-successes --with-ghc=$HC"
-(cd Cabal && timed $CMD) || exit 1
+CMD="$($CABALPLANLISTBIN Cabal-tests:test:unit-tests) $TESTSUITEJOBS --hide-successes --with-ghc=$HC"
+(cd Cabal-tests && timed $CMD) || exit 1
 
-CMD="$($CABALPLANLISTBIN Cabal:test:check-tests) $TESTSUITEJOBS --hide-successes"
-(cd Cabal && timed $CMD) || exit 1
+CMD="$($CABALPLANLISTBIN Cabal-tests:test:check-tests) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
 
-CMD="$($CABALPLANLISTBIN Cabal:test:parser-tests) $TESTSUITEJOBS --hide-successes"
-(cd Cabal && timed $CMD) || exit 1
+CMD="$($CABALPLANLISTBIN Cabal-tests:test:parser-tests) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
 
-CMD=$($CABALPLANLISTBIN Cabal:test:hackage-tests)
-(cd Cabal && timed $CMD read-fields) || exit 1
+CMD="$($CABALPLANLISTBIN Cabal-tests:test:rpmvercmp) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
 
+CMD="$($CABALPLANLISTBIN Cabal-tests:test:no-thunks-test) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
+
+CMD=$($CABALPLANLISTBIN Cabal-tests:test:hackage-tests)
+(cd Cabal-tests && timed $CMD read-fields) || exit 1
 if $HACKAGETESTSALL; then
-  (cd Cabal && timed $CMD parsec)    || exit 1
-  (cd Cabal && timed $CMD roundtrip) || exit 1
+    (cd Cabal-tests && timed $CMD parsec)    || exit 1
+    (cd Cabal-tests && timed $CMD roundtrip) || exit 1
 else
-  (cd Cabal && timed $CMD parsec d)    || exit 1
-  (cd Cabal && timed $CMD roundtrip k) || exit 1
+    (cd Cabal-tests && timed $CMD parsec d)    || exit 1
+    (cd Cabal-tests && timed $CMD roundtrip k) || exit 1
 fi
 }
 
@@ -430,7 +437,7 @@ step_cli_tests() {
 print_header "cabal-install: tests"
 
 # this are sorted in asc time used, quicker tests first.
-CMD="$($CABALPLANLISTBIN cabal-install:test:solver-quickcheck) $TESTSUITEJOBS --hide-successes"
+CMD="$($CABALPLANLISTBIN cabal-install:test:long-tests) $TESTSUITEJOBS --hide-successes"
 (cd cabal-install && timed $CMD) || exit 1
 
 # This doesn't work in parallel either
@@ -438,7 +445,7 @@ CMD="$($CABALPLANLISTBIN cabal-install:test:unit-tests) -j1 --hide-successes"
 (cd cabal-install && timed $CMD) || exit 1
 
 # Only single job, otherwise we fail with "Heap exhausted"
-CMD="$($CABALPLANLISTBIN cabal-install:test:memory-usage-tests) -j1 --hide-successes"
+CMD="$($CABALPLANLISTBIN cabal-install:test:mem-use-tests) -j1 --hide-successes"
 (cd cabal-install && timed $CMD) || exit 1
 
 # This test-suite doesn't like concurrency
@@ -452,7 +459,7 @@ CMD="$($CABALPLANLISTBIN cabal-install:test:integration-tests2) -j1 --hide-succe
 step_cli_suite() {
 print_header "cabal-install: cabal-testsuite"
 
-CMD="$($CABALPLANLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR --with-cabal=$($CABALPLANLISTBIN cabal-install:exe:cabal) $TESTSUITEJOBS --hide-successes"
+CMD="$($CABALPLANLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR --with-cabal=$($CABALPLANLISTBIN cabal-install:exe:cabal) $TESTSUITEJOBS  --with-ghc=$HC --hide-successes"
 (cd cabal-testsuite && timed $CMD) || exit 1
 }
 
@@ -481,7 +488,6 @@ for step in $STEPS; do
     case $step in
         print-config)             step_print_config            ;;
         print-tool-versions)      step_print_tool_versions     ;;
-        make-cabal-install-dev)   step_make_cabal_install_dev  ;;
         build)                    step_build                   ;;
         doctest)                  step_doctest                 ;;
         lib-tests)                step_lib_tests               ;;
@@ -492,7 +498,7 @@ for step in $STEPS; do
         solver-benchmarks-tests)  step_solver_benchmarks_tests ;;
         solver-benchmarks-run)    step_solver_benchmarks_run   ;;
         time-summary)             step_time_summary            ;;
-        *) 
+        *)
             echo "Invalid step $step"
             exit 1
             ;;
