@@ -12,6 +12,7 @@ module Data.Functor.Utils where
 import Data.Coerce (Coercible, coerce)
 import GHC.Base ( Applicative(..), Functor(..), Maybe(..), Monoid(..), Ord(..)
                 , Semigroup(..), ($), otherwise )
+import qualified GHC.List as List
 
 -- We don't expose Max and Min because, as Edward Kmett pointed out to me,
 -- there are two reasonable ways to define them. One way is to use Maybe, as we
@@ -34,6 +35,10 @@ instance Ord a => Semigroup (Max a) where
 -- | @since 4.8.0.0
 instance Ord a => Monoid (Max a) where
     mempty = Max Nothing
+    -- By default, we would get a lazy right fold. This forces the use of a strict
+    -- left fold instead.
+    mconcat = List.foldl' (<>) mempty
+    {-# INLINE mconcat #-}
 
 -- | @since 4.11.0.0
 instance Ord a => Semigroup (Min a) where
@@ -47,8 +52,12 @@ instance Ord a => Semigroup (Min a) where
 -- | @since 4.8.0.0
 instance Ord a => Monoid (Min a) where
     mempty = Min Nothing
+    -- By default, we would get a lazy right fold. This forces the use of a strict
+    -- left fold instead.
+    mconcat = List.foldl' (<>) mempty
+    {-# INLINE mconcat #-}
 
--- left-to-right state transformer
+-- left-to-right state-transforming monad
 newtype StateL s a = StateL { runStateL :: s -> (s, a) }
 
 -- | @since 4.0
@@ -67,7 +76,7 @@ instance Applicative (StateL s) where
             (s'', y) = ky s'
         in (s'', f x y)
 
--- right-to-left state transformer
+-- right-to-left state-transforming monad
 newtype StateR s a = StateR { runStateR :: s -> (s, a) }
 
 -- | @since 4.0
