@@ -7,7 +7,7 @@
 -- License     :  BSD-style (see the file libraries/base/LICENSE)
 -- 
 -- Maintainer  :  libraries@haskell.org
--- Stability   :  experimental
+-- Stability   :  stable
 -- Portability :  non-portable (uses Control.Monad.ST.Lazy)
 --
 -- Mutable references in the lazy ST monad.
@@ -23,9 +23,8 @@ module Data.STRef.Lazy (
         modifySTRef
  ) where
 
-import Control.Monad.ST.Lazy.Safe
+import Control.Monad.ST.Lazy
 import qualified Data.STRef as ST
-import Prelude
 
 newSTRef    :: a -> ST s (ST.STRef s a)
 readSTRef   :: ST.STRef s a -> ST s a
