@@ -8,7 +8,7 @@
 -- License     :  BSD-style (see the file libraries/base/LICENSE)
 --
 -- Maintainer  :  libraries@haskell.org
--- Stability   :  experimental
+-- Stability   :  stable
 -- Portability :  portable
 --
 -- Functions associated with the tuple data types.
@@ -16,7 +16,8 @@
 -----------------------------------------------------------------------------
 
 module Data.Tuple
-  ( fst
+  ( Solo (..)
+  , fst
   , snd
   , curry
   , uncurry
@@ -24,6 +25,7 @@ module Data.Tuple
   ) where
 
 import GHC.Base ()      -- Note [Depend on GHC.Tuple]
+import GHC.Tuple (Solo (..))
 
 default ()              -- Double isn't available yet
 
