@@ -26,7 +26,6 @@ module GHC.Err( absentErr, error, errorWithoutStackTrace, undefined ) where
 import GHC.Types (Char, RuntimeRep)
 import GHC.Stack.Types
 import GHC.Prim
-import GHC.Num.Integer ()   -- See Note [Depend on GHC.Num.Integer] in GHC.Base
 import {-# SOURCE #-} GHC.Exception
   ( errorCallWithCallStackException
   , errorCallException )
@@ -55,16 +54,12 @@ errorWithoutStackTrace s = raise# (errorCallException s)
 -- a partial stack trace, containing the call-site of each function
 -- with a HasCallStack constraint.
 --
--- In base, however, the only functions that have such constraints are
--- error and undefined, so the stack traces from partial functions in
--- base will never contain a call-site in user code. Instead we'll
--- usually just get the actual call to error. Base functions already
--- have a good habit of providing detailed error messages, including the
--- name of the offending partial function, so the partial stack-trace
--- does not provide any extra information, just noise. Thus, we export
--- the callstack-aware error, but within base we use the
--- errorWithoutStackTrace variant for more hygienic error messages.
-
+-- In base, error and undefined were the only functions that had such
+-- constraint. Errors like "Prelude.!!: negative index" are good, yet if the
+-- code base contains dozens of !! applications (including dependencies,
+-- which code is not as easily accessible), pinpointing the bad call is
+-- where the stack trace would help.  Therefore we annotate most calls to
+-- error, so users have a chance to get a better idea.
 
 -- | A special case of 'error'.
 -- It is expected that compilers will recognize this and insert error
@@ -72,7 +67,13 @@ errorWithoutStackTrace s = raise# (errorCallException s)
 -- appears.
 undefined :: forall (r :: RuntimeRep). forall (a :: TYPE r).
              HasCallStack => a
-undefined =  error "Prelude.undefined"
+-- This used to be
+--   undefined = error "Prelude.undefined"
+-- but that would add an extra call stack entry that is not actually helpful
+-- nor wanted (see #19886). We’d like to use withFrozenCallStack, but that
+-- is not available in this module yet, and making it so is hard. So let’s just
+-- use raise# directly.
+undefined = raise# (errorCallWithCallStackException "Prelude.undefined" ?callStack)
 
 -- | Used for compiler-generated error message;
 -- encoding saves bytes of string junk.
