@@ -1,11 +1,10 @@
+{-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE Trustworthy #-}
-{-# LANGUAGE BangPatterns
-           , CPP
-           , ExistentialQuantification
-           , NoImplicitPrelude
-           , TypeSynonymInstances
-           , FlexibleInstances
-  #-}
+
 -- TODO: use the new Windows IO manager
 module GHC.Event.TimerManager
     ( -- * Types
@@ -227,7 +226,7 @@ registerTimeout mgr us cb = do
 
 -- | Unregister an active timeout.
 unregisterTimeout :: TimerManager -> TimeoutKey -> IO ()
-unregisterTimeout mgr (TK key) = do
+unregisterTimeout mgr (TK key) =
   editTimeouts mgr (Q.delete key)
 
 -- | Update an active timeout to fire in the given number of
