@@ -2,6 +2,7 @@
 {-# LANGUAGE CPP
            , NoImplicitPrelude
            , BangPatterns
+           , RankNTypes
   #-}
 {-# OPTIONS_GHC -Wno-identities #-}
 -- Whether there are identities depends on the platform
@@ -22,7 +23,7 @@
 
 module GHC.IO.FD (
         FD(..),
-        openFile, mkFD, release,
+        openFileWith, openFile, mkFD, release,
         setNonBlockingMode,
         readRawBufferPtr, readRawBufferPtrNoBlock, writeRawBufferPtr,
         stdin, stdout, stderr
@@ -166,17 +167,46 @@ writeBuf' fd buf = do
 -- -----------------------------------------------------------------------------
 -- opening files
 
--- | Open a file and make an 'FD' for it.  Truncates the file to zero
--- size when the `IOMode` is `WriteMode`.
-openFile
+-- | Open a file and make an 'FD' for it. Truncates the file to zero size when
+-- the `IOMode` is `WriteMode`.
+--
+-- `openFileWith` takes two actions, @act1@ and @act2@, to perform after
+-- opening the file.
+--
+-- @act1@ is passed a file descriptor and I/O device type for the newly opened
+-- file. If an exception occurs in @act1@, then the file will be closed.
+-- @act1@ /must not/ close the file itself. If it does so and then receives an
+-- exception, then the exception handler will attempt to close it again, which
+-- is impermissable.
+--
+-- @act2@ is performed with asynchronous exceptions masked. It is passed a
+-- function to restore the masking state and the result of @act1@.  It /must
+-- not/ throw an exception (or deliver one via an interruptible operation)
+-- without first closing the file or arranging for it to be closed. @act2@
+-- /may/ close the file, but is not required to do so.  If @act2@ leaves the
+-- file open, then the file will remain open on return from `openFileWith`.
+--
+-- Code calling `openFileWith` that wishes to install a finalizer to close
+-- the file should do so in @act2@. Doing so in @act1@ could potentially close
+-- the file in the finalizer first and then in the exception handler. See
+-- 'GHC.IO.Handle.FD.openFile'' for an example of this use. Regardless, the
+-- caller is responsible for ensuring that the file is eventually closed,
+-- perhaps using 'Control.Exception.bracket'.
+
+openFileWith
   :: FilePath -- ^ file to open
   -> IOMode   -- ^ mode in which to open the file
   -> Bool     -- ^ open the file in non-blocking mode?
-  -> IO (FD,IODeviceType)
-
-openFile filepath iomode non_blocking =
+  -> (FD -> IODeviceType -> IO r) -- ^ @act1@: An action to perform
+                    -- on the file descriptor with the masking state
+                    -- restored and an exception handler that closes
+                    -- the file on exception.
+  -> ((forall x. IO x -> IO x) -> r -> IO s)
+                    -- ^ @act2@: An action to perform with async exceptions
+                    -- masked and no exception handler.
+  -> IO s
+openFileWith filepath iomode non_blocking act1 act2 =
   withFilePath filepath $ \ f ->
-
     let
       oflags1 = case iomode of
                   ReadMode      -> read_flags
@@ -195,25 +225,47 @@ openFile filepath iomode non_blocking =
       oflags | non_blocking = oflags2 .|. nonblock_flags
              | otherwise    = oflags2
     in do
+      -- We want to be sure all the arguments to c_interruptible_open
+      -- are fully evaluated *before* it slips under a mask (assuming we're
+      -- not already under a user-imposed mask).
+      oflags' <- evaluate oflags
+      -- NB. always use a safe open(), because we don't know whether open()
+      -- will be fast or not.  It can be slow on NFS and FUSE filesystems,
+      -- for example.
 
-    -- NB. always use a safe open(), because we don't know whether open()
-    -- will be fast or not.  It can be slow on NFS and FUSE filesystems,
-    -- for example.
-    fd <- throwErrnoIfMinus1Retry "openFile" $ c_safe_open f oflags 0o666
+      mask $ \restore -> do
+        fileno <- throwErrnoIfMinus1Retry "openFile" $
+                c_interruptible_open f oflags' 0o666
 
-    (fD,fd_type) <- mkFD fd iomode Nothing{-no stat-}
-                            False{-not a socket-}
-                            non_blocking
-            `catchAny` \e -> do _ <- c_close fd
-                                throwIO e
+        (fD,fd_type) <- mkFD fileno iomode Nothing{-no stat-}
+                                False{-not a socket-}
+                                non_blocking `onException` c_close fileno
 
-    -- we want to truncate() if this is an open in WriteMode, but only
-    -- if the target is a RegularFile.  ftruncate() fails on special files
-    -- like /dev/null.
-    when (iomode == WriteMode && fd_type == RegularFile) $
-      setSize fD 0
+        -- we want to truncate() if this is an open in WriteMode, but only
+        -- if the target is a RegularFile.  ftruncate() fails on special files
+        -- like /dev/null.
 
-    return (fD,fd_type)
+        when (iomode == WriteMode && fd_type == RegularFile) $
+          setSize fD 0 `onException` close fD
+
+        carry <- restore (act1 fD fd_type) `onException` close fD
+
+        act2 restore carry
+
+-- | Open a file and make an 'FD' for it.  Truncates the file to zero
+-- size when the `IOMode` is `WriteMode`. This function is difficult
+-- to use without potentially leaking the file descriptor on exception.
+-- In particular, it must be used with exceptions masked, which is a
+-- bit rude because the thread will be uninterruptible while the file
+-- path is being encoded. Use 'openFileWith' instead.
+openFile
+  :: FilePath -- ^ file to open
+  -> IOMode   -- ^ mode in which to open the file
+  -> Bool     -- ^ open the file in non-blocking mode?
+  -> IO (FD,IODeviceType)
+openFile filepath iomode non_blocking =
+  openFileWith filepath iomode non_blocking
+    (\ fd fd_type -> pure (fd, fd_type)) (\_ r -> pure r)
 
 std_flags, output_flags, read_flags, write_flags, rw_flags,
     append_flags, nonblock_flags :: CInt
@@ -379,8 +431,8 @@ getSize :: FD -> IO Integer
 getSize fd = fdFileSize (fdFD fd)
 
 setSize :: FD -> Integer -> IO ()
-setSize fd size = do
-  throwErrnoIf_ (/=0) "GHC.IO.FD.setSize"  $
+setSize fd size =
+  throwErrnoIf_ (/=0) "GHC.IO.FD.setSize" $
      c_ftruncate (fdFD fd) (fromIntegral size)
 
 devType :: FD -> IO IODeviceType
@@ -485,8 +537,8 @@ fdWriteNonBlocking fd ptr _offset bytes = do
 #if !defined(mingw32_HOST_OS)
 
 {-
-NOTE [nonblock]:
-
+Note [nonblock]
+~~~~~~~~~~~~~~~
 Unix has broken semantics when it comes to non-blocking I/O: you can
 set the O_NONBLOCK flag on an FD, but it applies to the all other FDs
 attached to the same underlying file, pipe or TTY; there's no way to
@@ -689,7 +741,7 @@ throwErrnoIfMinus1RetryOnBlock loc f on_block  =
         if err == eINTR
           then throwErrnoIfMinus1RetryOnBlock loc f on_block
           else if err == eWOULDBLOCK || err == eAGAIN
-                 then do on_block
+                 then on_block
                  else throwErrno loc
       else return res
 #endif
