@@ -5,8 +5,8 @@
            , BangPatterns
            , NondecreasingIndentation
            , MagicHash
+           , LambdaCase
   #-}
-{-# OPTIONS_GHC -Wno-name-shadowing #-}
 {-# OPTIONS_GHC -Wno-unused-matches #-}
 {-# OPTIONS_HADDOCK not-home #-}
 
@@ -28,11 +28,10 @@ module GHC.IO.Handle.Text (
         hWaitForInput, hGetChar, hGetLine, hGetContents, hPutChar, hPutStr,
         commitBuffer',       -- hack, see below
         hGetBuf, hGetBufSome, hGetBufNonBlocking, hPutBuf, hPutBufNonBlocking,
-        memcpy, hPutStrLn,
+        memcpy, hPutStrLn, hGetContents',
     ) where
 
 import GHC.IO
-import GHC.IO.FD
 import GHC.IO.Buffer
 import qualified GHC.IO.BufferedIO as Buffered
 import GHC.IO.Exception
@@ -46,8 +45,8 @@ import Foreign
 import Foreign.C
 
 import qualified Control.Exception as Exception
-import Data.Typeable
 import System.IO.Error
+import Data.Either (Either(..))
 import Data.Maybe
 
 import GHC.IORef
@@ -89,7 +88,7 @@ import GHC.List
 --
 
 hWaitForInput :: Handle -> Int -> IO Bool
-hWaitForInput h msecs = do
+hWaitForInput h msecs =
   wantReadableHandle_ "hWaitForInput" h $ \ handle_@Handle__{..} -> do
   cbuf <- readIORef haCharBuffer
 
@@ -187,8 +186,8 @@ hGetChar handle =
 
 hGetLine :: Handle -> IO String
 hGetLine h =
-  wantReadableHandle_ "hGetLine" h $ \ handle_ -> do
-     hGetLineBuffered handle_
+  wantReadableHandle_ "hGetLine" h $ \ handle_ ->
+    hGetLineBuffered handle_
 
 hGetLineBuffered :: Handle__ -> IO String
 hGetLineBuffered handle_@Handle__{..} = do
@@ -301,12 +300,12 @@ unpack_nl !buf !r !w acc0
               c <- peekElemOff pbuf i
               if (c == '\n' && i > r)
                  then do
-                         c1 <- peekElemOff pbuf (i-1)
-                         if (c1 == '\r')
-                            then unpackRB ('\n':acc) (i-2)
-                            else unpackRB ('\n':acc) (i-1)
-                 else do
-                         unpackRB (c : acc) (i-1)
+                   c1 <- peekElemOff pbuf (i-1)
+                   if (c1 == '\r')
+                      then unpackRB ('\n':acc) (i-2)
+                      else unpackRB ('\n':acc) (i-1)
+                 else
+                   unpackRB (c : acc) (i-1)
      in do
      c <- peekElemOff pbuf (w-1)
      if (c == '\r')
@@ -321,7 +320,7 @@ unpack_nl !buf !r !w acc0
                 return (str, w)
 
 -- Note [#5536]
---
+-- ~~~~~~~~~~~~
 -- We originally had
 --
 --    let c' = desurrogatifyRoundtripCharacter c in
@@ -439,20 +438,105 @@ getSomeCharacters handle_@Handle__{..} buf@Buffer{..} =
     1 | haInputNL == CRLF -> do
       (c,_) <- readCharBuf bufRaw bufL
       if c == '\r'
-         then do -- shuffle the '\r' to the beginning.  This is only safe
-                 -- if we're about to call readTextDevice, otherwise it
-                 -- would mess up flushCharBuffer.
-                 -- See [note Buffer Flushing], GHC.IO.Handle.Types
-                 _ <- writeCharBuf bufRaw 0 '\r'
-                 let buf' = buf{ bufL=0, bufR=1 }
-                 readTextDevice handle_ buf'
-         else do
-                 return buf
+      then do
+        -- shuffle the '\r' to the beginning.  This is only safe
+        -- if we're about to call readTextDevice, otherwise it
+        -- would mess up flushCharBuffer.
+        -- See [note Buffer Flushing], GHC.IO.Handle.Types
+        _ <- writeCharBuf bufRaw 0 '\r'
+        let buf' = buf{ bufL=0, bufR=1 }
+        readTextDevice handle_ buf'
+      else
+        return buf
 
     -- buffer has some chars in it already: just return it
     _otherwise ->
       return buf
 
+-- -----------------------------------------------------------------------------
+-- hGetContents'
+
+-- We read everything into a list of CharBuffer chunks, and convert it lazily
+-- to a string, which minimizes memory usage.
+-- In the worst case, space usage is at most that of the complete String,
+-- as the chunks can be garbage collected progressively.
+-- For streaming consumers, space usage is at most that of the list of chunks.
+
+-- | The 'hGetContents'' operation reads all input on the given handle
+-- before returning it as a 'String' and closing the handle.
+--
+-- @since 4.15.0.0
+
+hGetContents' :: Handle -> IO String
+hGetContents' handle = do
+    es <- wantReadableHandle "hGetContents'" handle (strictRead handle)
+    case es of
+      Right s -> return s
+      Left e ->
+          case fromException e of
+            Just ioe -> throwIO (augmentIOError ioe "hGetContents'" handle)
+            Nothing -> throwIO e
+
+strictRead :: Handle -> Handle__ -> IO (Handle__, Either SomeException String)
+strictRead h handle_@Handle__{..} = do
+    cbuf <- readIORef haCharBuffer
+    cbufs <- strictReadLoop' handle_ [] cbuf
+    (handle_', me) <- hClose_help handle_
+    case me of
+      Just e -> return (handle_', Left e)
+      Nothing -> do
+        s <- lazyBuffersToString haInputNL cbufs ""
+        return (handle_', Right s)
+
+strictReadLoop :: Handle__ -> [CharBuffer] -> CharBuffer -> IO [CharBuffer]
+strictReadLoop handle_ cbufs cbuf0 = do
+    mcbuf <- Exception.catch
+        (do r <- readTextDevice handle_ cbuf0
+            return (Just r))
+        (\e -> if isEOFError e
+                  then return Nothing
+                  else throw e)
+    case mcbuf of
+      Nothing -> return (cbuf0 : cbufs)
+      Just cbuf1 -> strictReadLoop' handle_ cbufs cbuf1
+
+-- If 'cbuf' is full, allocate a new buffer.
+strictReadLoop' :: Handle__ -> [CharBuffer] -> CharBuffer -> IO [CharBuffer]
+strictReadLoop' handle_ cbufs cbuf
+    | isFullCharBuffer cbuf = do
+        cbuf' <- newCharBuffer dEFAULT_CHAR_BUFFER_SIZE ReadBuffer
+        strictReadLoop handle_ (cbuf : cbufs) cbuf'
+    | otherwise = strictReadLoop handle_ cbufs cbuf
+
+-- Lazily convert a list of buffers to a String. The buffers are
+-- in reverse order: the first buffer is the end of the String.
+lazyBuffersToString :: Newline -> [CharBuffer] -> String -> IO String
+lazyBuffersToString LF = loop where
+    loop [] s = return s
+    loop (Buffer{..} : cbufs) s = do
+        s' <- unsafeInterleaveIO (unpack bufRaw bufL bufR s)
+        loop cbufs s'
+lazyBuffersToString CRLF = loop '\0' where
+    loop before [] s = return s
+    loop before (Buffer{..} : cbufs) s
+        | bufL == bufR = loop before cbufs s  -- skip empty buffers
+        | otherwise = do
+            -- When a CRLF is broken across two buffers, we already have a newline
+            -- from decoding the LF, so we ignore the CR in the current buffer.
+            s1 <- if before == '\n'
+                     then return s
+                     else do
+                       -- We restore trailing CR not followed by LF.
+                       c <- peekCharBuf bufRaw (bufR - 1)
+                       if c == '\r'
+                          then return ('\r' : s)
+                          else return s
+            s2 <- unsafeInterleaveIO (do
+                (s2, _) <- unpack_nl bufRaw bufL bufR s1
+                return s2)
+            c0 <- peekCharBuf bufRaw bufL
+            loop c0 cbufs s2
+
 -- ---------------------------------------------------------------------------
 -- hPutChar
 
@@ -467,35 +551,34 @@ getSomeCharacters handle_@Handle__{..} buf@Buffer{..} =
 --  * 'isPermissionError' if another system resource limit would be exceeded.
 
 hPutChar :: Handle -> Char -> IO ()
-hPutChar handle c = do
-    c `seq` return ()
-    wantWritableHandle "hPutChar" handle $ \ handle_  -> do
-     hPutcBuffered handle_ c
+hPutChar handle !c =
+    wantWritableHandle "hPutChar" handle $ \ handle_  ->
+      hPutcBuffered handle_ c
 
 hPutcBuffered :: Handle__ -> Char -> IO ()
 hPutcBuffered handle_@Handle__{..} c = do
   buf <- readIORef haCharBuffer
   if c == '\n'
      then do buf1 <- if haOutputNL == CRLF
-                        then do
-                          buf1 <- putc buf '\r'
-                          putc buf1 '\n'
-                        else do
-                          putc buf '\n'
+                     then do
+                       buf1 <- putc buf '\r'
+                       putc buf1 '\n'
+                     else
+                       putc buf '\n'
              writeCharBuffer handle_ buf1
-             when is_line $ flushByteWriteBuffer handle_
+             when isLine $ flushByteWriteBuffer handle_
       else do
           buf1 <- putc buf c
           writeCharBuffer handle_ buf1
           return ()
   where
-    is_line = case haBufferMode of
+    isLine = case haBufferMode of
                 LineBuffering -> True
                 _             -> False
 
-    putc buf@Buffer{ bufRaw=raw, bufR=w } c = do
+    putc buf@Buffer{ bufRaw=raw, bufR=w } c' = do
        debugIO ("putc: " ++ summaryBuffer buf)
-       w'  <- writeCharBuf raw w c
+       w'  <- writeCharBuf raw w c'
        return buf{ bufR = w' }
 
 -- ---------------------------------------------------------------------------
@@ -533,13 +616,15 @@ hPutStr handle str = hPutStr' handle str False
 -- | The same as 'hPutStr', but adds a newline character.
 hPutStrLn :: Handle -> String -> IO ()
 hPutStrLn handle str = hPutStr' handle str True
-  -- An optimisation: we treat hPutStrLn specially, to avoid the
-  -- overhead of a single putChar '\n', which is quite high now that we
-  -- have to encode eagerly.
 
 {-# NOINLINE hPutStr' #-}
 hPutStr' :: Handle -> String -> Bool -> IO ()
 hPutStr' handle str add_nl =
+  -- An optimisation: hPutStr' takes an additional "add_nl" boolean parameter to
+  -- implement hPutStrLn efficiently. When LineBuffering or BlockBuffering modes
+  -- are enabled, it performs the encoding of the string and of the new-line
+  -- character(s) in the same loop, directly flushing buffers appropriately at
+  -- the end if LineBuffering mode is used.
   do
     (buffer_mode, nl) <-
          wantWritableHandle "hPutStr" handle $ \h_ -> do
@@ -550,20 +635,18 @@ hPutStr' handle str add_nl =
        (NoBuffering, _) -> do
             hPutChars handle str        -- v. slow, but we don't care
             when add_nl $ hPutChar handle '\n'
-       (LineBuffering, buf) -> do
+       (LineBuffering, buf) ->
             writeBlocks handle True  add_nl nl buf str
-       (BlockBuffering _, buf) -> do
+       (BlockBuffering _, buf) ->
             writeBlocks handle False add_nl nl buf str
 
 hPutChars :: Handle -> [Char] -> IO ()
 hPutChars _      [] = return ()
 hPutChars handle (c:cs) = hPutChar handle c >> hPutChars handle cs
 
+-- Buffer offset is always zero.
 getSpareBuffer :: Handle__ -> IO (BufferMode, CharBuffer)
-getSpareBuffer Handle__{haCharBuffer=ref,
-                        haBuffers=spare_ref,
-                        haBufferMode=mode}
- = do
+getSpareBuffer Handle__{haCharBuffer=ref, haBuffers=spare_ref, haBufferMode=mode} =
    case mode of
      NoBuffering -> return (mode, errorWithoutStackTrace "no buffer!")
      _ -> do
@@ -581,71 +664,73 @@ getSpareBuffer Handle__{haCharBuffer=ref,
 -- NB. performance-critical code: eyeball the Core.
 writeBlocks :: Handle -> Bool -> Bool -> Newline -> Buffer CharBufElem -> String -> IO ()
 writeBlocks hdl line_buffered add_nl nl
-            buf@Buffer{ bufRaw=raw, bufSize=len } s =
-  let
-   shoveString :: Int -> [Char] -> [Char] -> IO ()
-   shoveString !n [] [] = do
-        commitBuffer hdl raw len n False{-no flush-} True{-release-}
-   shoveString !n [] rest = do
-        shoveString n rest []
-   shoveString !n (c:cs) rest
-     -- n+1 so we have enough room to write '\r\n' if necessary
-     | n + 1 >= len = do
+            buf@Buffer{ bufRaw=raw, bufSize=len } s = shoveString 0 s
+  where
+   {-# INLINE new_line #-} -- we don't want to allocate a closure for it
+   new_line !n = do
+     n1 <- case nl of
+            CRLF -> writeCharBuf raw n '\r'
+            _    -> pure n
+     n2 <- writeCharBuf raw n1 '\n'
+     if line_buffered
+        then do
+          -- end of line, so write and flush
+          commitBuffer hdl raw len n2 True{-flush-} False
+          pure 0
+        else
+          pure n2
+
+   shoveString !n = \case
+    [] -> if add_nl
+            then do
+              n' <- new_line n
+              commitBuffer hdl raw len n' False{-no flush-} True{-release-}
+            else
+              commitBuffer hdl raw len n False{-no flush-} True{-release-}
+
+    -- n+1 so we have enough room to write '\r\n' if necessary
+    cs | n + 1 >= len -> do
         commitBuffer hdl raw len n False{-flush-} False
-        shoveString 0 (c:cs) rest
-     | c == '\n'  =  do
-        n' <- if nl == CRLF
-                 then do
-                    n1 <- writeCharBuf raw n  '\r'
-                    writeCharBuf raw n1 '\n'
-                 else do
-                    writeCharBuf raw n c
-        if line_buffered
-           then do
-                -- end of line, so write and flush
-               commitBuffer hdl raw len n' True{-flush-} False
-               shoveString 0 cs rest
-           else do
-               shoveString n' cs rest
-     | otherwise = do
+        shoveString 0 cs
+
+    ('\n':cs) -> do
+        n' <- new_line n
+        shoveString n' cs
+
+    (c:cs) -> do
         n' <- writeCharBuf raw n c
-        shoveString n' cs rest
-  in
-  shoveString 0 s (if add_nl then "\n" else "")
+        shoveString n' cs
 
 -- -----------------------------------------------------------------------------
 -- commitBuffer handle buf sz count flush release
 --
 -- Write the contents of the buffer 'buf' ('sz' bytes long, containing
 -- 'count' bytes of data) to handle (handle must be block or line buffered).
-
-commitBuffer
-        :: Handle                       -- handle to commit to
-        -> RawCharBuffer -> Int         -- address and size (in bytes) of buffer
-        -> Int                          -- number of bytes of data in buffer
-        -> Bool                         -- True <=> flush the handle afterward
-        -> Bool                         -- release the buffer?
-        -> IO ()
-
+commitBuffer :: Handle                       -- handle to commit to
+             -> RawCharBuffer -> Int         -- address and size (in bytes) of buffer
+             -> Int                          -- number of bytes of data in buffer
+             -> Bool                         -- True <=> flush the handle afterward
+             -> Bool                         -- release the buffer?
+             -> IO ()
 commitBuffer hdl !raw !sz !count flush release =
   wantWritableHandle "commitBuffer" hdl $ \h_@Handle__{..} -> do
-      debugIO ("commitBuffer: sz=" ++ show sz ++ ", count=" ++ show count
-            ++ ", flush=" ++ show flush ++ ", release=" ++ show release)
-
-      writeCharBuffer h_ Buffer{ bufRaw=raw, bufState=WriteBuffer,
-                                 bufL=0, bufR=count, bufSize=sz }
-
-      when flush $ flushByteWriteBuffer h_
-
-      -- release the buffer if necessary
-      when release $ do
-          -- find size of current buffer
-          old_buf@Buffer{ bufSize=size } <- readIORef haCharBuffer
-          when (sz == size) $ do
-               spare_bufs <- readIORef haBuffers
-               writeIORef haBuffers (BufferListCons raw spare_bufs)
-
-      return ()
+    let debugMsg = ("commitBuffer: sz=" ++ show sz ++ ", count=" ++ show count
+                    ++ ", flush=" ++ show flush ++ ", release=" ++ show release
+                    ++ ", handle=" ++ show hdl)
+    debugIO debugMsg
+      -- Offset taken from handle
+    writeCharBuffer h_ Buffer{ bufRaw=raw, bufState=WriteBuffer, bufOffset=0,
+                               bufL=0, bufR=count, bufSize=sz }
+    when flush $ flushByteWriteBuffer h_
+    -- release the buffer if necessary
+    when release $ do
+      -- find size of current buffer
+      old_buf@Buffer{ bufSize=size } <- readIORef haCharBuffer
+      when (sz == size) $ do
+        spare_bufs <- readIORef haBuffers
+        writeIORef haBuffers (BufferListCons raw spare_bufs)
+    -- bb <- readIORef haByteBuffer
+    -- debugIO ("commitBuffer: buffer=" ++ summaryBuffer bb ++ ", handle=" ++ show hdl)
 
 -- backwards compatibility; the text package uses this
 commitBuffer' :: RawCharBuffer -> Int -> Int -> Bool -> Bool -> Handle__
@@ -656,7 +741,7 @@ commitBuffer' raw sz@(I# _) count@(I# _) flush release h_@Handle__{..}
             ++ ", flush=" ++ show flush ++ ", release=" ++ show release)
 
       let this_buf = Buffer{ bufRaw=raw, bufState=WriteBuffer,
-                             bufL=0, bufR=count, bufSize=sz }
+                             bufL=0, bufR=count, bufSize=sz, bufOffset=0 }
 
       writeCharBuffer h_ this_buf
 
@@ -727,67 +812,84 @@ hPutBuf' handle ptr count can_block
           -- it is set to LineBuffering, be conservative and flush
           -- anyway (we didn't check for newlines in the data).
           case haBufferMode of
-             BlockBuffering _      -> do return ()
-             _line_or_no_buffering -> do flushWriteBuffer h_
+             BlockBuffering _      -> return ()
+             _line_or_no_buffering -> flushWriteBuffer h_
           return r
 
+-- TODO: Possible optimisation:
+--       If we know that `w + count > size`, we should write both the
+--       handle buffer and the `ptr` in a single `writev()` syscall.
 bufWrite :: Handle__-> Ptr Word8 -> Int -> Bool -> IO Int
-bufWrite h_@Handle__{..} ptr count can_block =
-  seq count $ do  -- strictness hack
-  old_buf@Buffer{ bufRaw=old_raw, bufR=w, bufSize=size }
-     <- readIORef haByteBuffer
+bufWrite h_@Handle__{..} ptr !count can_block = do
+  -- Get buffer to determine size and free space in buffer
+  old_buf@Buffer{ bufR=w, bufSize=size }
+      <- readIORef haByteBuffer
 
-  -- TODO: Possible optimisation:
-  --       If we know that `w + count > size`, we should write both the
-  --       handle buffer and the `ptr` in a single `writev()` syscall.
-
-  -- Need to buffer and enough room in handle buffer?
-  -- There's no need to buffer if the data to be written is larger than
+  -- There's no need to buffer if the incoming data is larger than
   -- the handle buffer (`count >= size`).
-  if (count < size && count <= size - w)
-        -- We need to buffer and there's enough room in the buffer:
-        -- just copy the data in and update bufR.
-        then do debugIO ("hPutBuf: copying to buffer, w=" ++ show w)
-                copyToRawBuffer old_raw w ptr count
-                let copied_buf = old_buf{ bufR = w + count }
-                -- If the write filled the buffer completely, we need to flush,
-                -- to maintain the "INVARIANTS on Buffers" from
-                -- GHC.IO.Buffer.checkBuffer: "a write buffer is never full".
-                if (count == size - w)
-                  then do
-                    debugIO "hPutBuf: flushing full buffer after writing"
-                    flushed_buf <- Buffered.flushWriteBuffer haDevice copied_buf
-                            -- TODO: we should do a non-blocking flush here
-                    writeIORef haByteBuffer flushed_buf
-                  else do
-                    writeIORef haByteBuffer copied_buf
-                return count
+  -- Check if we can try to buffer the given chunk of data.
+  b <- if (count < size && count <= size - w)
+        then bufferChunk h_ old_buf ptr count
+        else do
+          -- The given data does not fit into the buffer.
+          -- Either because it's too large for the buffer
+          -- or the buffer is too full. Either way we need
+          -- to flush the buffered data first.
+          flushed_buf <- flushByteWriteBufferGiven h_ old_buf
+          if count < size
+              -- The data is small enough to be buffered.
+              then bufferChunk h_ flushed_buf ptr count
+              else do
+                let offset = bufOffset flushed_buf
+                !bytes <- if can_block
+                            then writeChunk            h_ (castPtr ptr) offset count
+                            else writeChunkNonBlocking h_ (castPtr ptr) offset count
+                -- Update buffer with actual bytes written.
+                writeIORef haByteBuffer $! bufferAddOffset bytes flushed_buf
+                return bytes
+  debugIO "hPutBuf: done"
+  return b
 
-        -- else, we have to flush any existing handle buffer data
-        -- and can then write out the data in `ptr` directly.
-        else do -- No point flushing when there's nothing in the buffer.
-                when (w > 0) $ do
-                  debugIO "hPutBuf: flushing first"
-                  flushed_buf <- Buffered.flushWriteBuffer haDevice old_buf
-                          -- TODO: we should do a non-blocking flush here
-                  writeIORef haByteBuffer flushed_buf
-                -- if we can fit in the buffer, then just loop
-                if count < size
-                   then bufWrite h_ ptr count can_block
-                   else if can_block
-                           then do writeChunk h_ (castPtr ptr) count
-                                   return count
-                           else writeChunkNonBlocking h_ (castPtr ptr) count
+-- Flush the given buffer via the handle, return the flushed buffer
+flushByteWriteBufferGiven :: Handle__ -> Buffer Word8 -> IO (Buffer Word8)
+flushByteWriteBufferGiven h_@Handle__{..} bbuf =
+  if (not (isEmptyBuffer bbuf))
+    then do
+      bbuf' <- Buffered.flushWriteBuffer haDevice bbuf
+      debugIO ("flushByteWriteBufferGiven: bbuf=" ++ summaryBuffer bbuf')
+      writeIORef haByteBuffer bbuf'
+      return bbuf'
+    else
+      return bbuf
 
-writeChunk :: Handle__ -> Ptr Word8 -> Int -> IO ()
-writeChunk h_@Handle__{..} ptr bytes
-  | Just fd <- cast haDevice  =  RawIO.write (fd::FD) ptr bytes
-  | otherwise = error "Todo: hPutBuf"
+-- Fill buffer and return bytes buffered/written.
+-- Flushes buffer if it's full after adding the data.
+bufferChunk :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> IO Int
+bufferChunk h_@Handle__{..} old_buf@Buffer{ bufRaw=raw, bufR=w, bufSize=size } ptr !count = do
+    debugIO ("hPutBuf: copying to buffer, w=" ++ show w)
+    copyToRawBuffer raw w ptr count
+    let copied_buf = old_buf{ bufR = w + count }
+    -- If the write filled the buffer completely, we need to flush,
+    -- to maintain the "INVARIANTS on Buffers" from
+    -- GHC.IO.Buffer.checkBuffer: "a write buffer is never full".
+    if isFullBuffer copied_buf
+      then do
+        -- TODO: we should do a non-blocking flush here
+        debugIO "hPutBuf: flushing full buffer after writing"
+        _ <- flushByteWriteBufferGiven h_ copied_buf
+        return ()
+      else
+        writeIORef haByteBuffer copied_buf
+    return count
 
-writeChunkNonBlocking :: Handle__ -> Ptr Word8 -> Int -> IO Int
-writeChunkNonBlocking h_@Handle__{..} ptr bytes
-  | Just fd <- cast haDevice  =  RawIO.writeNonBlocking (fd::FD) ptr bytes
-  | otherwise = error "Todo: hPutBuf"
+writeChunk :: Handle__ -> Ptr Word8 -> Word64 -> Int -> IO Int
+writeChunk h_@Handle__{..} ptr offset bytes
+  = do RawIO.write haDevice ptr offset bytes
+       return bytes
+
+writeChunkNonBlocking :: Handle__ -> Ptr Word8 -> Word64 -> Int -> IO Int
+writeChunkNonBlocking h_@Handle__{..} ptr offset bytes
+  = RawIO.writeNonBlocking haDevice ptr offset bytes
 
 -- ---------------------------------------------------------------------------
 -- hGetBuf
@@ -813,12 +915,16 @@ hGetBuf h !ptr count
   | count <  0 = illegalBufferSize h "hGetBuf" count
   | otherwise =
       wantReadableHandle_ "hGetBuf" h $ \ h_@Handle__{..} -> do
-         flushCharReadBuffer h_
-         buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
+          debugIO $ ":: hGetBuf - " ++ show h ++ " - " ++ show count
+          flushCharReadBuffer h_
+          buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
             <- readIORef haByteBuffer
-         if isEmptyBuffer buf
-            then bufReadEmpty    h_ buf (castPtr ptr) 0 count
-            else bufReadNonEmpty h_ buf (castPtr ptr) 0 count
+          debugIO ("hGetBuf: " ++ summaryBuffer buf)
+          res <- if isEmptyBuffer buf
+                    then bufReadEmpty    h_ buf (castPtr ptr) 0 count
+                    else bufReadNonEmpty h_ buf (castPtr ptr) 0 count
+          debugIO "** hGetBuf done."
+          return res
 
 -- small reads go through the buffer, large reads are satisfied by
 -- taking data first from the buffer and then direct from the file
@@ -826,9 +932,14 @@ hGetBuf h !ptr count
 
 bufReadNonEmpty :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> Int -> IO Int
 bufReadNonEmpty h_@Handle__{..}
+                -- w for width, r for ... read ptr?
                 buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
                 ptr !so_far !count
  = do
+        debugIO ":: bufReadNonEmpty"
+        -- We use < instead of <= because for count == avail
+        -- we need to reset bufL and bufR to zero.
+        -- See also: INVARIANTS on Buffers
         let avail = w - r
         if (count < avail)
            then do
@@ -844,30 +955,47 @@ bufReadNonEmpty h_@Handle__{..}
             so_far' = so_far + avail
             ptr' = ptr `plusPtr` avail
 
-        if remaining == 0
+        debugIO ("bufReadNonEmpty: " ++ summaryBuffer buf' ++ " s:" ++ show so_far' ++ " r:" ++ show remaining)
+        b <- if remaining == 0
            then return so_far'
            else bufReadEmpty h_ buf' ptr' so_far' remaining
+        debugIO ":: bufReadNonEmpty - done"
+        return b
 
-
+-- We want to read more data, but the buffer is empty. (buffL == buffR == 0)
+-- See also Note [INVARIANTS on Buffers] in Buffer.hs
 bufReadEmpty :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> Int -> IO Int
 bufReadEmpty h_@Handle__{..}
-             buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
+             buf@Buffer{ bufRaw=raw, bufR=w, bufL=_r, bufSize=sz, bufOffset=bff }
              ptr so_far count
- | count > sz, Just fd <- cast haDevice = loop fd 0 count
+ | count > sz
+ = do
+        bytes_read <- loop haDevice 0 bff count
+        -- bytes_read includes so_far (content that was in the buffer)
+        -- but that is already accounted for in the old offset, so don't
+        -- count it twice.
+        let buf1 = bufferAddOffset (fromIntegral $ bytes_read - so_far) buf
+        writeIORef haByteBuffer buf1
+        debugIO ("bufReadEmpty1.1: " ++ summaryBuffer buf1 ++ " read:" ++ show bytes_read)
+        return bytes_read
  | otherwise = do
-     (r,buf') <- Buffered.fillReadBuffer haDevice buf
-     if r == 0
-        then return so_far
-        else do writeIORef haByteBuffer buf'
-                bufReadNonEmpty h_ buf' ptr so_far count
+        (r,buf') <- Buffered.fillReadBuffer haDevice buf
+        writeIORef haByteBuffer buf'
+        if r == 0 -- end of file reached
+            then return so_far
+            else bufReadNonEmpty h_ buf' ptr so_far count
  where
-  loop :: FD -> Int -> Int -> IO Int
-  loop fd off bytes | bytes <= 0 = return (so_far + off)
-  loop fd off bytes = do
-    r <- RawIO.read (fd::FD) (ptr `plusPtr` off) bytes
+  -- Read @bytes@ byte into ptr. Repeating the read until either zero
+  -- bytes where read, or we are done reading.
+  loop :: RawIO.RawIO dev => dev -> Int -> Word64 -> Int -> IO Int
+  loop dev delta off bytes | bytes <= 0 = return (so_far + delta)
+  loop dev delta off bytes = do
+    r <- RawIO.read dev (ptr `plusPtr` delta) off bytes
+    debugIO $ show ptr ++ " - loop read@" ++ show delta ++ ": " ++ show r
+    debugIO $ "next:" ++ show (delta + r) ++ " - left:" ++ show (bytes - r)
     if r == 0
-        then return (so_far + off)
-        else loop fd (off + r) (bytes - r)
+        then return (so_far + delta)
+        else loop dev (delta + r) (off + fromIntegral r) (bytes - r)
 
 -- ---------------------------------------------------------------------------
 -- hGetBufSome
@@ -896,10 +1024,13 @@ hGetBufSome h !ptr count
   | otherwise =
       wantReadableHandle_ "hGetBufSome" h $ \ h_@Handle__{..} -> do
          flushCharReadBuffer h_
-         buf@Buffer{ bufSize=sz } <- readIORef haByteBuffer
+         buf@Buffer{ bufSize=sz, bufOffset=offset } <- readIORef haByteBuffer
          if isEmptyBuffer buf
             then case count > sz of  -- large read? optimize it with a little special case:
-                    True | Just fd <- haFD h_ -> do RawIO.read fd (castPtr ptr) count
+                    True -> do bytes <- RawIO.read haDevice (castPtr ptr) offset count
+                               -- Update buffer with actual bytes written.
+                               writeIORef haByteBuffer $! bufferAddOffset bytes buf
+                               return bytes
                     _ -> do (r,buf') <- Buffered.fillReadBuffer haDevice buf
                             if r == 0
                                then return 0
@@ -912,9 +1043,6 @@ hGetBufSome h !ptr count
               let count' = min count (bufferElems buf)
               in bufReadNBNonEmpty h_ buf (castPtr ptr) 0 count'
 
-haFD :: Handle__ -> Maybe FD
-haFD h_@Handle__{..} = cast haDevice
-
 -- | 'hGetBufNonBlocking' @hdl buf count@ reads data from the handle @hdl@
 -- into the buffer @buf@ until either EOF is reached, or
 -- @count@ 8-bit bytes have been read, or there is no more data available
@@ -949,25 +1077,27 @@ hGetBufNonBlocking h !ptr count
 
 bufReadNBEmpty :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> Int -> IO Int
 bufReadNBEmpty   h_@Handle__{..}
-                 buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
+                 buf@Buffer{ bufRaw=raw, bufR=w, bufL=_r, bufSize=sz
+                           , bufOffset=offset }
                  ptr so_far count
-  | count > sz,
-    Just fd <- cast haDevice = do
-       m <- RawIO.readNonBlocking (fd::FD) ptr count
+  | count > sz = do
+       m <- RawIO.readNonBlocking haDevice ptr offset count
        case m of
          Nothing -> return so_far
-         Just n  -> return (so_far + n)
+         Just n  -> do -- Update buffer with actual bytes written.
+                       writeIORef haByteBuffer $! bufferAddOffset n buf
+                       return (so_far + n)
 
  | otherwise = do
-     buf <- readIORef haByteBuffer
+    --  buf <- readIORef haByteBuffer
      (r,buf') <- Buffered.fillReadBuffer0 haDevice buf
      case r of
        Nothing -> return so_far
        Just 0  -> return so_far
-       Just r  -> do
+       Just r'  -> do
          writeIORef haByteBuffer buf'
-         bufReadNBNonEmpty h_ buf' ptr so_far (min count r)
-                          -- NOTE: new count is    min count r
+         bufReadNBNonEmpty h_ buf' ptr so_far (min count r')
+                          -- NOTE: new count is    min count r'
                           -- so we will just copy the contents of the
                           -- buffer in the recursive call, and not
                           -- loop again.
@@ -979,6 +1109,9 @@ bufReadNBNonEmpty h_@Handle__{..}
                   ptr so_far count
   = do
         let avail = w - r
+        -- We use < instead of <= because for count == avail
+        -- we need to reset bufL and bufR to zero.
+        -- See also [INVARIANTS on Buffers] in Buffer.hs
         if (count < avail)
            then do
                 copyFromRawBuffer ptr raw r count
