@@ -1,6 +1,5 @@
 {-# LANGUAGE Trustworthy #-}
 {-# LANGUAGE DataKinds #-}
-{-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE TypeOperators #-}
 {-# LANGUAGE FlexibleInstances #-}
@@ -11,14 +10,13 @@
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE MagicHash #-}
-{-# LANGUAGE PolyKinds #-}
+{-# LANGUAGE TypeApplications #-}
 
 {-|
-
 GHC's @DataKinds@ language extension lifts data constructors, natural
 numbers, and strings to the type level. This module provides the
-primitives needed for working with type-level numbers (the 'Nat' kind)
-and strings (the 'Symbol') kind. It also defines the 'TypeError' type
+primitives needed for working with type-level numbers (the 'Nat' kind),
+strings (the 'Symbol' kind), and characters (the 'Char' kind). It also defines the 'TypeError' type
 family, a feature that makes use of type-level strings to support user
 defined type errors.
 
@@ -33,21 +31,26 @@ working with type-level data will be defined in a separate library.
 
 module GHC.TypeLits
   ( -- * Kinds
-    Nat, Symbol  -- Both declared in GHC.Types in package ghc-prim
+    N.Natural, N.Nat, Symbol  -- Symbol is declared in GHC.Types in package ghc-prim
 
     -- * Linking type and value level
   , N.KnownNat, natVal, natVal'
   , KnownSymbol, symbolVal, symbolVal'
-  , N.SomeNat(..), SomeSymbol(..)
-  , someNatVal, someSymbolVal
-  , N.sameNat, sameSymbol
+  , KnownChar, charVal, charVal'
+  , N.SomeNat(..), SomeSymbol(..), SomeChar(..)
+  , someNatVal, someSymbolVal, someCharVal
+  , N.sameNat, sameSymbol, sameChar
+  , OrderingI(..)
+  , N.cmpNat, cmpSymbol, cmpChar
 
 
     -- * Functions on type literals
   , type (N.<=), type (N.<=?), type (N.+), type (N.*), type (N.^), type (N.-)
   , type N.Div, type N.Mod, type N.Log2
   , AppendSymbol
-  , N.CmpNat, CmpSymbol
+  , N.CmpNat, CmpSymbol, CmpChar
+  , ConsSymbol, UnconsSymbol
+  , CharToNat, NatToChar
 
   -- * User-defined type errors
   , TypeError
@@ -55,20 +58,21 @@ module GHC.TypeLits
 
   ) where
 
-import GHC.Base(Eq(..), Ord(..), Ordering(..), otherwise)
-import GHC.Types( Nat, Symbol )
+import GHC.Base(Eq(..), Ord(..), Ordering(..), String, otherwise, withDict)
+import GHC.Types(Symbol, Char)
+import GHC.TypeError(ErrorMessage(..), TypeError)
 import GHC.Num(Integer, fromInteger)
-import GHC.Base(String)
 import GHC.Show(Show(..))
 import GHC.Read(Read(..))
 import GHC.Real(toInteger)
-import GHC.Prim(magicDict, Proxy#)
+import GHC.Prim(Proxy#)
 import Data.Maybe(Maybe(..))
 import Data.Proxy (Proxy(..))
 import Data.Type.Equality((:~:)(Refl))
+import Data.Type.Ord(OrderingI(..))
 import Unsafe.Coerce(unsafeCoerce)
 
-import GHC.TypeNats (KnownNat)
+import GHC.TypeLits.Internal(CmpSymbol, CmpChar)
 import qualified GHC.TypeNats as N
 
 --------------------------------------------------------------------------------
@@ -81,7 +85,7 @@ class KnownSymbol (n :: Symbol) where
   symbolSing :: SSymbol n
 
 -- | @since 4.7.0.0
-natVal :: forall n proxy. KnownNat n => proxy n -> Integer
+natVal :: forall n proxy. N.KnownNat n => proxy n -> Integer
 natVal p = toInteger (N.natVal p)
 
 -- | @since 4.7.0.0
@@ -90,7 +94,7 @@ symbolVal _ = case symbolSing :: SSymbol n of
                 SSymbol x -> x
 
 -- | @since 4.8.0.0
-natVal' :: forall n. KnownNat n => Proxy# n -> Integer
+natVal' :: forall n. N.KnownNat n => Proxy# n -> Integer
 natVal' p = toInteger (N.natVal' p)
 
 -- | @since 4.8.0.0
@@ -103,6 +107,20 @@ symbolVal' _ = case symbolSing :: SSymbol n of
 data SomeSymbol = forall n. KnownSymbol n => SomeSymbol (Proxy n)
                   -- ^ @since 4.7.0.0
 
+-- | @since 4.16.0.0
+class KnownChar (n :: Char) where
+  charSing :: SChar n
+
+charVal :: forall n proxy. KnownChar n => proxy n -> Char
+charVal _ = case charSing :: SChar n of
+                 SChar x -> x
+
+charVal' :: forall n. KnownChar n => Proxy# n -> Char
+charVal' _ = case charSing :: SChar n of
+                SChar x -> x
+
+data SomeChar = forall n. KnownChar n => SomeChar (Proxy n)
+
 -- | Convert an integer into an unknown type-level natural.
 --
 -- @since 4.7.0.0
@@ -136,65 +154,55 @@ instance Show SomeSymbol where
 instance Read SomeSymbol where
   readsPrec p xs = [ (someSymbolVal a, ys) | (a,ys) <- readsPrec p xs ]
 
---------------------------------------------------------------------------------
 
--- | Comparison of type-level symbols, as a function.
+-- | Convert a character into an unknown type-level char.
 --
--- @since 4.7.0.0
-type family CmpSymbol (m :: Symbol) (n :: Symbol) :: Ordering
+-- @since 4.16.0.0
+someCharVal :: Char -> SomeChar
+someCharVal n   = withSChar SomeChar (SChar n) Proxy
+{-# NOINLINE someCharVal #-}
+
+instance Eq SomeChar where
+  SomeChar x == SomeChar y = charVal x == charVal y
+
+instance Ord SomeChar where
+  compare (SomeChar x) (SomeChar y) = compare (charVal x) (charVal y)
+
+instance Show SomeChar where
+  showsPrec p (SomeChar x) = showsPrec p (charVal x)
+
+instance Read SomeChar where
+  readsPrec p xs = [ (someCharVal a, ys) | (a,ys) <- readsPrec p xs ]
+
+--------------------------------------------------------------------------------
 
 -- | Concatenation of type-level symbols.
 --
 -- @since 4.10.0.0
 type family AppendSymbol (m ::Symbol) (n :: Symbol) :: Symbol
 
--- | A description of a custom type error.
-data {-kind-} ErrorMessage = Text Symbol
-                             -- ^ Show the text as is.
+-- Char-related type families
 
-                           | forall t. ShowType t
-                             -- ^ Pretty print the type.
-                             -- @ShowType :: k -> ErrorMessage@
-
-                           | ErrorMessage :<>: ErrorMessage
-                             -- ^ Put two pieces of error message next
-                             -- to each other.
-
-                           | ErrorMessage :$$: ErrorMessage
-                             -- ^ Stack two pieces of error message on top
-                             -- of each other.
-
-infixl 5 :$$:
-infixl 6 :<>:
-
--- | The type-level equivalent of 'Prelude.error'.
+-- | Extending a type-level symbol with a type-level character
 --
--- The polymorphic kind of this type allows it to be used in several settings.
--- For instance, it can be used as a constraint, e.g. to provide a better error
--- message for a non-existent instance,
---
--- @
--- -- in a context
--- instance TypeError (Text "Cannot 'Show' functions." :$$:
---                     Text "Perhaps there is a missing argument?")
---       => Show (a -> b) where
---     showsPrec = error "unreachable"
--- @
---
--- It can also be placed on the right-hand side of a type-level function
--- to provide an error for an invalid case,
---
--- @
--- type family ByteSize x where
---    ByteSize Word16   = 2
---    ByteSize Word8    = 1
---    ByteSize a        = TypeError (Text "The type " :<>: ShowType a :<>:
---                                   Text " is not exportable.")
--- @
---
--- @since 4.9.0.0
-type family TypeError (a :: ErrorMessage) :: b where
+-- @since 4.16.0.0
+type family ConsSymbol (a :: Char) (b :: Symbol) :: Symbol
 
+-- | This type family yields type-level `Just` storing the first character
+-- of a symbol and its tail if it is defined and `Nothing` otherwise.
+--
+-- @since 4.16.0.0
+type family UnconsSymbol (a :: Symbol) :: Maybe (Char, Symbol)
+
+-- | Convert a character to its Unicode code point (cf. `Data.Char.ord`)
+--
+-- @since 4.16.0.0
+type family CharToNat (c :: Char) :: N.Nat
+
+-- | Convert a Unicode code point to a character (cf. `Data.Char.chr`)
+--
+-- @since 4.16.0.0
+type family NatToChar (n :: N.Nat) :: Char
 
 --------------------------------------------------------------------------------
 
@@ -208,14 +216,61 @@ sameSymbol x y
   | symbolVal x == symbolVal y  = Just (unsafeCoerce Refl)
   | otherwise                   = Nothing
 
+
+-- | We either get evidence that this function was instantiated with the
+-- same type-level characters, or 'Nothing'.
+--
+-- @since 4.16.0.0
+sameChar :: (KnownChar a, KnownChar b) =>
+              proxy1 a -> proxy2 b -> Maybe (a :~: b)
+sameChar x y
+  | charVal x == charVal y  = Just (unsafeCoerce Refl)
+  | otherwise                = Nothing
+
+-- | Like 'sameSymbol', but if the symbols aren't equal, this additionally
+-- provides proof of LT or GT.
+--
+-- @since 4.16.0.0
+cmpSymbol :: forall a b proxy1 proxy2. (KnownSymbol a, KnownSymbol b)
+          => proxy1 a -> proxy2 b -> OrderingI a b
+cmpSymbol x y = case compare (symbolVal x) (symbolVal y) of
+  EQ -> case unsafeCoerce (Refl, Refl) :: (CmpSymbol a b :~: 'EQ, a :~: b) of
+    (Refl, Refl) -> EQI
+  LT -> case unsafeCoerce Refl :: (CmpSymbol a b :~: 'LT) of
+    Refl -> LTI
+  GT -> case unsafeCoerce Refl :: (CmpSymbol a b :~: 'GT) of
+    Refl -> GTI
+
+-- | Like 'sameChar', but if the Chars aren't equal, this additionally
+-- provides proof of LT or GT.
+--
+-- @since 4.16.0.0
+cmpChar :: forall a b proxy1 proxy2. (KnownChar a, KnownChar b)
+        => proxy1 a -> proxy2 b -> OrderingI a b
+cmpChar x y = case compare (charVal x) (charVal y) of
+  EQ -> case unsafeCoerce (Refl, Refl) :: (CmpChar a b :~: 'EQ, a :~: b) of
+    (Refl, Refl) -> EQI
+  LT -> case unsafeCoerce Refl :: (CmpChar a b :~: 'LT) of
+    Refl -> LTI
+  GT -> case unsafeCoerce Refl :: (CmpChar a b :~: 'GT) of
+    Refl -> GTI
+
+
 --------------------------------------------------------------------------------
 -- PRIVATE:
 
 newtype SSymbol (s :: Symbol) = SSymbol String
 
-data WrapS a b = WrapS (KnownSymbol a => Proxy a -> b)
-
--- See Note [magicDictId magic] in "basicType/MkId.hs"
-withSSymbol :: (KnownSymbol a => Proxy a -> b)
+-- See Note [withDict] in "GHC.Tc.Instance.Class" in GHC
+withSSymbol :: forall a b.
+               (KnownSymbol a => Proxy a -> b)
             -> SSymbol a      -> Proxy a -> b
-withSSymbol f x y = magicDict (WrapS f) x y
+withSSymbol f x y = withDict @(KnownSymbol a) x f y
+
+newtype SChar (s :: Char) = SChar Char
+
+-- See Note [withDict] in "GHC.Tc.Instance.Class" in GHC
+withSChar :: forall a b.
+             (KnownChar a => Proxy a -> b)
+            -> SChar a      -> Proxy a -> b
+withSChar f x y = withDict @(KnownChar a) x f y
