@@ -9,14 +9,13 @@ import Data.Monoid (mempty)
 import qualified Data.Set as Set
 import IntSetValidity (valid)
 import Prelude hiding (lookup, null, map, filter, foldr, foldl)
-import Test.Framework
-import Test.Framework.Providers.HUnit
-import Test.Framework.Providers.QuickCheck2
-import Test.HUnit hiding (Test, Testable)
-import Test.QuickCheck hiding ((.&.))
+import Test.Tasty
+import Test.Tasty.HUnit
+import Test.Tasty.QuickCheck hiding ((.&.))
 
 main :: IO ()
-main = defaultMain [ testCase "lookupLT" test_lookupLT
+main = defaultMain $ testGroup "intset-properties"
+                   [ testCase "lookupLT" test_lookupLT
                    , testCase "lookupGT" test_lookupGT
                    , testCase "lookupLE" test_lookupLE
                    , testCase "lookupGE" test_lookupGE
@@ -71,6 +70,9 @@ main = defaultMain [ testCase "lookupLT" test_lookupLT
                    , testProperty "prop_splitRoot" prop_splitRoot
                    , testProperty "prop_partition" prop_partition
                    , testProperty "prop_filter" prop_filter
+                   , testProperty "takeWhileAntitone" prop_takeWhileAntitone
+                   , testProperty "dropWhileAntitone" prop_dropWhileAntitone
+                   , testProperty "spanAntitone" prop_spanAntitone
                    , testProperty "prop_bitcount" prop_bitcount
                    , testProperty "prop_alterF_list" prop_alterF_list
                    , testProperty "prop_alterF_const" prop_alterF_const
@@ -421,6 +423,26 @@ prop_filter s i =
      valid evens .&&.
      parts === (odds, evens)
 
+prop_takeWhileAntitone :: Int -> [Int] -> Property
+prop_takeWhileAntitone x ys =
+  let l = takeWhileAntitone (<x) (fromList ys)
+  in  valid l .&&.
+      l === fromList (List.filter (<x) ys)
+
+prop_dropWhileAntitone :: Int -> [Int] -> Property
+prop_dropWhileAntitone x ys =
+  let r = dropWhileAntitone (<x) (fromList ys)
+  in  valid r .&&.
+      r === fromList (List.filter (>=x) ys)
+
+prop_spanAntitone :: Int -> [Int] -> Property
+prop_spanAntitone x ys =
+  let (l, r) = spanAntitone (<x) (fromList ys)
+  in  valid l .&&.
+      valid r .&&.
+      l === fromList (List.filter (<x) ys) .&&.
+      r === fromList (List.filter (>=x) ys)
+
 prop_bitcount :: Int -> Word -> Bool
 prop_bitcount a w = bitcount_orig a w == bitcount_new a w
   where
