@@ -3,18 +3,21 @@
 {-# LANGUAGE DefaultSignatures #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE GADTs #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE Safe #-}
+{-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE TypeOperators #-}
-
-#if __GLASGOW_HASKELL__ >= 706
 {-# LANGUAGE PolyKinds #-}
+{-# LANGUAGE EmptyCase #-}
+
+#if __GLASGOW_HASKELL__ >= 811 && __GLASGOW_HASKELL__ < 901
+-- For the Option instance (https://gitlab.haskell.org/ghc/ghc/issues/15028)
+{-# OPTIONS_GHC -Wno-deprecations #-}
 #endif
 
-#if __GLASGOW_HASKELL__ >= 708
-{-# LANGUAGE EmptyCase #-}
-#endif
+#define BYTEARRAY_IN_BASE (__GLASGOW_HASKELL__ >= 903)
+-- At the moment of writing GHC source tree has not yet bumped `base` version,
+-- so using __GLASGOW_HASKELL__ as a proxy instead of MIN_VERSION_base(4,17,0).
 
 -----------------------------------------------------------------------------
 -- |
@@ -104,69 +107,54 @@ import Foreign.Ptr
 import Foreign.C.Types
 import System.Exit ( ExitCode(..) )
 import System.Mem.StableName ( StableName )
-
-#if MIN_VERSION_base(4,6,0)
 import Data.Ord ( Down(Down) )
-#else
-import Control.DeepSeq.BackDoor ( Down(Down) )
-#endif
-
-#if MIN_VERSION_base(4,7,0)
 import Data.Proxy ( Proxy(Proxy) )
-#endif
 
 #if MIN_VERSION_base(4,10,0)
 import Data.Type.Equality ( (:~:), (:~~:) )
-#elif MIN_VERSION_base(4,9,0)
+#else
 import Data.Type.Equality ( (:~:) )
-#elif MIN_VERSION_base(4,7,0)
-import Control.DeepSeq.BackDoor ( (:~:) )
 #endif
 
-#if MIN_VERSION_base(4,8,0)
 import Data.Functor.Identity ( Identity(..) )
 import Data.Typeable ( rnfTypeRep, rnfTyCon )
 import Data.Void ( Void, absurd )
 import Numeric.Natural ( Natural )
-#else
-import Data.Typeable ( typeRepTyCon, typeRepArgs, tyConPackage, tyConModule, tyConName )
-#endif
 
-#if MIN_VERSION_base(4,9,0)
 import Data.List.NonEmpty ( NonEmpty (..) )
 import Data.Semigroup as Semi
-#endif
 
-#if MIN_VERSION_base(4,9,0)
 import GHC.Stack.Types ( CallStack(..), SrcLoc(..) )
 import Data.Functor.Compose
 import qualified Data.Functor.Sum as Functor
 import qualified Data.Functor.Product as Functor
-#elif MIN_VERSION_base(4,8,1)
-import GHC.Stack ( CallStack(..) )
-import GHC.SrcLoc ( SrcLoc(..) )
-#endif
 
 import GHC.Fingerprint.Type ( Fingerprint(..) )
 import GHC.Generics
 
+#ifdef MIN_VERSION_ghc_prim
+#if MIN_VERSION_ghc_prim(0,7,0)
+import GHC.Tuple (Solo (..))
+#endif
+#endif
+
+#if BYTEARRAY_IN_BASE
+import Data.Array.Byte (ByteArray(..), MutableByteArray(..))
+#endif
+
 -- | Hidden internal type-class
 class GNFData arity f where
   grnf :: RnfArgs arity a -> f a -> ()
 
 instance GNFData arity V1 where
-#if __GLASGOW_HASKELL__ >= 708
   grnf _ x = case x of {}
-#else
-  grnf _ !_ = error "Control.DeepSeq.rnf: uninhabited type"
-#endif
 
 data Zero
 data One
 
-data RnfArgs arity a where
-  RnfArgs0 :: RnfArgs Zero a
-  RnfArgs1  :: (a -> ()) -> RnfArgs One a
+data family RnfArgs arity a
+data instance RnfArgs Zero a = RnfArgs0
+newtype instance RnfArgs One a = RnfArgs1 (a -> ())
 
 instance GNFData arity U1 where
   grnf _ U1 = ()
@@ -179,17 +167,11 @@ instance GNFData arity a => GNFData arity (M1 i c a) where
   grnf args = grnf args . unM1
   {-# INLINEABLE grnf #-}
 
-#if MIN_VERSION_base(4,9,0)
-  -- | Because 'URec' did not exist prior to @base-4.9@, this instance is only
-  -- defined on @base-4.9@ or later.
-  --
-  -- @since 1.4.5.0
 instance GNFData arity (URec a) where
   grnf _ = rwhnf -- Every URec data instance consists of a single data
                  -- constructor containing a single strict field, so reducing
                  -- any URec instance to WHNF suffices to reduce it to NF.
   {-# INLINEABLE grnf #-}
-#endif
 
 instance (GNFData arity a, GNFData arity b) => GNFData arity (a :*: b) where
   grnf args (x :*: y) = grnf args x `seq` grnf args y
@@ -210,6 +192,7 @@ instance (NFData1 f, GNFData One g) => GNFData One (f :.: g) where
     grnf args = liftRnf (grnf args) . unComp1
 
 infixr 0 $!!
+infixr 0 `deepseq`
 
 -- | 'deepseq': fully evaluates the first argument, before returning the
 -- second.
@@ -447,7 +430,6 @@ instance NFData Word64   where rnf = rwhnf
 -- | @since 1.4.4.0
 instance NFData MaskingState where rnf = rwhnf
 
-#if MIN_VERSION_base(4,7,0)
 -- |@since 1.4.0.0
 instance NFData (Proxy a) where rnf Proxy = ()
 -- |@since 1.4.3.0
@@ -459,7 +441,6 @@ instance NFData (a :~: b) where rnf = rwhnf
 instance NFData1 ((:~:) a) where liftRnf _ = rwhnf
 -- | @since 1.4.3.0
 instance NFData2 (:~:) where liftRnf2 _ _ = rwhnf
-#endif
 
 #if MIN_VERSION_base(4,10,0)
 -- | @since 1.4.3.0
@@ -470,7 +451,6 @@ instance NFData1 ((:~~:) a) where liftRnf _ = rwhnf
 instance NFData2 (:~~:) where liftRnf2 _ _ = rwhnf
 #endif
 
-#if MIN_VERSION_base(4,8,0)
 -- |@since 1.4.0.0
 instance NFData a => NFData (Identity a) where
     rnf = rnf1
@@ -487,7 +467,6 @@ instance NFData Void where
 
 -- |@since 1.4.0.0
 instance NFData Natural  where rnf = rwhnf
-#endif
 
 -- |@since 1.3.0.0
 instance NFData (Fixed a) where rnf = rwhnf
@@ -502,7 +481,6 @@ instance NFData (a -> b) where rnf = rwhnf
 
 --Rational and complex numbers.
 
-#if MIN_VERSION_base(4,9,0)
 -- | Available on @base >=4.9@
 --
 -- @since 1.4.3.0
@@ -535,9 +513,6 @@ instance (NFData1 f, NFData1 g, NFData a) => NFData (Functor.Product f g a) wher
   rnf = rnf1
 
 instance NFData a => NFData (Ratio a) where
-#else
-instance (Integral a, NFData a) => NFData (Ratio a) where
-#endif
   rnf x = rnf (numerator x, denominator x)
 
 instance (NFData a) => NFData (Complex a) where
@@ -589,27 +564,16 @@ instance NFData2 Const where
 
 -- We should use MIN_VERSION array(0,5,1,1) but that's not possible.
 -- There isn't an underscore to not break C preprocessor
-#if __GLASGOW_HASKELL__ >= 711
 instance (NFData a, NFData b) => NFData (Array a b) where
-#else
-instance (Ix a, NFData a, NFData b) => NFData (Array a b) where
-#endif
     rnf x = rnf (bounds x, Data.Array.elems x)
 
-#if __GLASGOW_HASKELL__ >= 711
 -- |@since 1.4.3.0
 instance (NFData a) => NFData1 (Array a) where
-#else
--- |@since 1.4.3.0
-instance (Ix a, NFData a) => NFData1 (Array a) where
-#endif
     liftRnf r x = rnf (bounds x) `seq` liftRnf r (Data.Array.elems x)
 
-#if __GLASGOW_HASKELL__ >= 711
 -- |@since 1.4.3.0
 instance NFData2 Array where
     liftRnf2 r r' x = liftRnf2 r r (bounds x) `seq` liftRnf r' (Data.Array.elems x)
-#endif
 
 -- |@since 1.4.0.0
 instance NFData a => NFData (Down a) where rnf = rnf1
@@ -668,7 +632,6 @@ instance NFData ThreadId where
 instance NFData Unique where
     rnf = rwhnf -- assumes `newtype Unique = Unique Integer`
 
-#if MIN_VERSION_base(4,8,0)
 -- | __NOTE__: Prior to @deepseq-1.4.4.0@ this instance was only defined for @base-4.8.0.0@ and later.
 --
 -- @since 1.4.0.0
@@ -680,21 +643,7 @@ instance NFData TypeRep where
 -- @since 1.4.0.0
 instance NFData TyCon where
     rnf tycon = rnfTyCon tycon
-#else
--- | __NOTE__: Prior to @deepseq-1.4.4.0@ this instance was only defined for @base-4.8.0.0@ and later.
---
--- @since 1.4.0.0
-instance NFData TypeRep where
-    rnf tr = rnf (typeRepTyCon tr) `seq` rnf (typeRepArgs tr)
 
--- | __NOTE__: Prior to @deepseq-1.4.4.0@ this instance was only defined for @base-4.8.0.0@ and later.
---
--- @since 1.4.0.0
-instance NFData TyCon where
-    rnf tc = rnf (tyConPackage tc) `seq`
-             rnf (tyConModule  tc) `seq`
-             rnf (tyConName    tc)
-#endif
 
 -- | __NOTE__: Only strict in the reference and not the referenced value.
 --
@@ -857,7 +806,6 @@ instance NFData ExitCode where
 ----------------------------------------------------------------------------
 -- instances previously provided by semigroups package
 
-#if MIN_VERSION_base(4,9,0)
 -- |@since 1.4.2.0
 instance NFData a => NFData (NonEmpty a) where rnf = rnf1
 -- |@since 1.4.3.0
@@ -902,6 +850,7 @@ instance NFData m => NFData (WrappedMonoid m) where rnf = rnf1
 instance NFData1 WrappedMonoid where
   liftRnf r (WrapMonoid a) = r a
 
+#if __GLASGOW_HASKELL__ < 901
 -- |@since 1.4.2.0
 instance NFData a => NFData (Option a) where rnf = rnf1
 -- |@since 1.4.3.0
@@ -912,7 +861,6 @@ instance NFData1 Option where
 ----------------------------------------------------------------------------
 -- GHC.Stack
 
-#if MIN_VERSION_base(4,9,0)
 -- |@since 1.4.2.0
 instance NFData SrcLoc where
   rnf (SrcLoc a b c d e f g) = rnf a `seq` rnf b `seq` rnf c `seq`
@@ -924,26 +872,20 @@ instance NFData CallStack where
   rnf (PushCallStack a b c) = rnf a `seq` rnf b `seq` rnf c
   rnf (FreezeCallStack a)   = rnf a
 
-#elif MIN_VERSION_base(4,8,1)
--- |@since 1.4.2.0
-instance NFData SrcLoc where
-  -- base-4.8 didn't expose the 'SrcLoc' constructor
-  rnf sl = rnf (srcLocPackage   sl) `seq`
-           rnf (srcLocModule    sl) `seq`
-           rnf (srcLocFile      sl) `seq`
-           rnf (srcLocStartLine sl) `seq`
-           rnf (srcLocStartCol  sl) `seq`
-           rnf (srcLocEndLine   sl) `seq`
-           rnf (srcLocEndCol    sl)
-
--- |@since 1.4.2.0
-instance NFData CallStack where
-  rnf = rnf . getCallStack
-#endif
-
 ----------------------------------------------------------------------------
 -- Tuples
 
+#ifdef MIN_VERSION_ghc_prim
+#if MIN_VERSION_ghc_prim(0,7,0)
+-- |@since 1.4.6.0
+instance NFData a => NFData (Solo a) where
+  rnf (Solo a) = rnf a
+-- |@since 1.4.6.0
+instance NFData1 Solo where
+  liftRnf r (Solo a) = r a
+#endif
+#endif
+
 instance (NFData a, NFData b) => NFData (a,b) where rnf = rnf2
 -- |@since 1.4.3.0
 instance (NFData a) => NFData1 ((,) a) where liftRnf = liftRnf2 rnf
@@ -1021,3 +963,16 @@ instance (NFData a1, NFData a2, NFData a3, NFData a4, NFData a5, NFData a6, NFDa
 instance (NFData a1, NFData a2, NFData a3, NFData a4, NFData a5, NFData a6, NFData a7) =>
          NFData2 ((,,,,,,,,) a1 a2 a3 a4 a5 a6 a7) where
   liftRnf2 r r' (x1,x2,x3,x4,x5,x6,x7,x8,x9) = rnf x1 `seq` rnf x2 `seq` rnf x3 `seq` rnf x4 `seq` rnf x5 `seq` rnf x6 `seq` rnf x7 `seq` r x8 `seq` r' x9
+
+----------------------------------------------------------------------------
+-- ByteArray
+
+#if BYTEARRAY_IN_BASE
+-- |@since 1.4.7.0
+instance NFData ByteArray where
+  rnf (ByteArray _) = ()
+
+-- |@since 1.4.8.0
+instance NFData (MutableByteArray s) where
+  rnf (MutableByteArray _) = ()
+#endif
