@@ -86,7 +86,7 @@ qualify pw str
     | otherwise = str
     where
         prelude = ["elem","uncurry","snd","fst","not","null","if","then","else"
-                  ,"True","False","Just","Nothing","fromJust","concat","isPrefixOf","isSuffixOf","any"]
+                  ,"True","False","Just","Nothing","fromJust","concat","isPrefixOf","isSuffixOf","any","foldr"]
         fpops = ["</>","<.>","-<.>"]
 
 
