@@ -8,6 +8,14 @@ import System.Console.Haskeline.Completion(Completion)
 
 import Control.Concurrent
 import Control.Concurrent.STM
+import Control.Exception (Exception, SomeException(..))
+import Control.Monad.Catch
+    ( MonadMask
+    , bracket
+    , handle
+    , throwM
+    , finally
+    )
 import Data.Word
 import Control.Exception (fromException, AsyncException(..))
 import Data.Typeable
@@ -17,7 +25,7 @@ import System.IO.Error (isEOFError)
 import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as BC
 
-class (MonadReader Layout m, MonadException m) => Term m where
+class (MonadReader Layout m, MonadIO m, MonadMask m) => Term m where
     reposition :: Layout -> LineChars -> m ()
     moveToNextLine :: LineChars -> m ()
     printLines :: [String] -> m ()
@@ -34,7 +42,7 @@ data RunTerm = RunTerm {
             -- | Write unicode characters to stdout.
             putStrOut :: String -> IO (),
             termOps :: Either TermOps FileOps,
-            wrapInterrupt :: forall a . IO a -> IO a,
+            wrapInterrupt :: forall a m . (MonadIO m, MonadMask m) => m a -> m a,
             closeTerm :: IO ()
     }
 
@@ -68,7 +76,7 @@ flushEventQueue print' eventChan = yield >> loopUntilFlushed
 -- Backends can assume that getLocaleLine, getLocaleChar and maybeReadNewline
 -- are "wrapped" by wrapFileInput.
 data FileOps = FileOps {
-            withoutInputEcho :: forall m a . MonadException m => m a -> m a,
+            withoutInputEcho :: forall m a . (MonadIO m, MonadMask m) => m a -> m a,
             -- ^ Perform an action without echoing input.
             wrapFileInput :: forall a . IO a -> IO a,
             getLocaleLine :: MaybeT IO String,
@@ -100,12 +108,12 @@ instance Exception Interrupt where
 
 
 
-class (MonadReader Prefs m , MonadReader Layout m, MonadException m)
+class (MonadReader Prefs m , MonadReader Layout m, MonadIO m, MonadMask m)
         => CommandMonad m where
     runCompletion :: (String,String) -> m (String,[Completion])
 
 instance {-# OVERLAPPABLE #-} (MonadTrans t, CommandMonad m, MonadReader Prefs (t m),
-        MonadException (t m),
+        MonadIO (t m), MonadMask (t m),
         MonadReader Layout (t m))
             => CommandMonad (t m) where
     runCompletion = lift . runCompletion
@@ -153,14 +161,14 @@ data Layout = Layout {width, height :: Int}
 -- Utility functions for the various backends.
 
 -- | Utility function since we're not using the new IO library yet.
-hWithBinaryMode :: MonadException m => Handle -> m a -> m a
+hWithBinaryMode :: (MonadIO m, MonadMask m) => Handle -> m a -> m a
 hWithBinaryMode h = bracket (liftIO $ hGetEncoding h)
                         (maybe (return ()) (liftIO . hSetEncoding h))
                         . const . (liftIO (hSetBinaryMode h True) >>)
 
 -- | Utility function for changing a property of a terminal for the duration of
 -- a computation.
-bracketSet :: MonadException m => IO a -> (a -> IO ()) -> a -> m b -> m b
+bracketSet :: (MonadMask m, MonadIO m) => IO a -> (a -> IO ()) -> a -> m b -> m b
 bracketSet getState set newState f = bracket (liftIO getState)
                             (liftIO . set)
                             (\_ -> liftIO (set newState) >> f)
@@ -193,10 +201,10 @@ hMaybeReadNewline h = returnOnEOF () $ do
         c <- hLookAhead h
         when (c == '\n') $ getChar >> return ()
 
-returnOnEOF :: MonadException m => a -> m a -> m a
+returnOnEOF :: MonadMask m => a -> m a -> m a
 returnOnEOF x = handle $ \e -> if isEOFError e
                                 then return x
-                                else throwIO e
+                                else throwM e
 
 -- | Utility function to correctly get a line of input as an undecoded ByteString.
 hGetLocaleLine :: Handle -> MaybeT IO ByteString
