@@ -6,6 +6,8 @@
 #endif
 {-# LANGUAGE InterruptibleFFI #-}
 
+#include <ghcplatform.h>
+
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  System.Process
@@ -20,7 +22,7 @@
 --
 -----------------------------------------------------------------------------
 
--- ToDo:
+-- TODO:
 --      * Flag to control whether exiting the parent also kills the child.
 
 module System.Process (
@@ -49,6 +51,7 @@ module System.Process (
     showCommandForUser,
     Pid,
     getPid,
+    getCurrentPid,
 
     -- ** Control-C handling on Unix
     -- $ctlc-handling
@@ -93,13 +96,19 @@ import System.IO
 import System.IO.Error (mkIOError, ioeSetErrorString)
 
 #if defined(WINDOWS)
-import System.Win32.Process (getProcessId, ProcessId)
+import System.Win32.Process (getProcessId, getCurrentProcessId, ProcessId)
 #else
+import System.Posix.Process (getProcessID)
 import System.Posix.Types (CPid (..))
 #endif
 
 import GHC.IO.Exception ( ioException, IOErrorType(..), IOException(..) )
 
+#if defined(wasm32_HOST_ARCH)
+import GHC.IO.Exception ( unsupportedOperation )
+import System.IO.Error
+#endif
+
 -- | The platform specific type for a process identifier.
 --
 -- This is always an integral type. Width and signedness are platform specific.
@@ -200,6 +209,13 @@ always the desired behavior. In cases where you would like to leave the
 instead. All created @Handle@s are initially in text mode; if you need them
 to be in binary mode then use 'hSetBinaryMode'.
 
+@/ph/@ contains a handle to the running process.  On Windows
+'use_process_jobs' can be set in CreateProcess in order to create a
+Win32 Job object to monitor a process tree's progress.  If it is set
+then that job is also returned inside @/ph/@.  @/ph/@ can be used to
+kill all running sub-processes.  This feature has been available since
+1.5.0.0.
+
 -}
 createProcess
   :: CreateProcess
@@ -249,10 +265,10 @@ withCreateProcess_ fun c action =
                      (\(m_in, m_out, m_err, ph) -> action m_in m_out m_err ph)
 
 -- | Cleans up the process.
--- 
--- This function is meant to be invoked from any application level cleanup 
+--
+-- This function is meant to be invoked from any application level cleanup
 -- handler. It terminates the process, and closes any 'CreatePipe' 'handle's.
--- 
+--
 -- @since 1.6.4.0
 cleanupProcess :: (Maybe Handle, Maybe Handle, Maybe Handle, ProcessHandle)
                -> IO ()
@@ -374,7 +390,7 @@ processFailedException fun cmd args exit_code =
 -- @SIGINT@ to every process using the console. The standard solution is that
 -- while running an interactive program, ignore @SIGINT@ in the parent, and let
 -- it be handled in the child process. If that process then terminates due to
--- the @SIGINT@ signal, then at that point treat it as if we had recieved the
+-- the @SIGINT@ signal, then at that point treat it as if we had received the
 -- @SIGINT@ ourselves and begin an orderly shutdown.
 --
 -- This behaviour is implemented by 'createProcess' (and
@@ -645,15 +661,39 @@ getPid (ProcessHandle mh _ _) = do
     _ -> return Nothing
 
 
+-- ----------------------------------------------------------------------------
+-- getCurrentPid
+
+-- | Returns the PID (process ID) of the current process. On POSIX systems,
+-- this calls 'getProcessID' from "System.Posix.Process" in the @unix@ package.
+-- On Windows, this calls 'getCurrentProcessId' from "System.Win32.Process" in
+-- the @Win32@ package.
+--
+-- @since 1.6.12.0
+getCurrentPid :: IO Pid
+getCurrentPid =
+#ifdef WINDOWS
+    getCurrentProcessId
+#else
+    getProcessID
+#endif
+
+
 -- ----------------------------------------------------------------------------
 -- waitForProcess
 
 {- | Waits for the specified process to terminate, and returns its exit code.
+On Unix systems, may throw 'UserInterrupt' when using 'delegate_ctlc'.
 
 GHC Note: in order to call @waitForProcess@ without blocking all the
 other threads in the system, you must compile the program with
 @-threaded@.
 
+Note that it is safe to call @waitForProcess@ for the same process in multiple
+threads. When the process ends, threads blocking on this call will wake in
+FIFO order. When using 'delegate_ctlc' and the process is interrupted, only
+the first waiting thread will throw 'UserInterrupt'.
+
 (/Since: 1.2.0.0/) On Unix systems, a negative value @'ExitFailure' -/signum/@
 indicates that the child was terminated by signal @/signum/@.
 The signal numbers are platform-specific, so to test for a specific signal use
@@ -672,15 +712,17 @@ waitForProcess ph@(ProcessHandle _ delegating_ctlc _) = lockWaitpid $ do
     OpenHandle h  -> do
         -- don't hold the MVar while we call c_waitForProcess...
         e <- waitForProcess' h
-        e' <- modifyProcessHandle ph $ \p_' ->
+        (e', was_open) <- modifyProcessHandle ph $ \p_' ->
           case p_' of
-            ClosedHandle e' -> return (p_', e')
+            ClosedHandle e' -> return (p_', (e', False))
             OpenExtHandle{} -> fail "waitForProcess(OpenExtHandle): this cannot happen"
             OpenHandle ph'  -> do
               closePHANDLE ph'
-              when delegating_ctlc $
-                endDelegateControlC e
-              return (ClosedHandle e, e)
+              return (ClosedHandle e, (e, True))
+        -- endDelegateControlC after closing the handle, since it
+        -- may throw UserInterrupt
+        when (was_open && delegating_ctlc) $
+          endDelegateControlC e
         return e'
 #if defined(WINDOWS)
     OpenExtHandle h job -> do
@@ -694,9 +736,8 @@ waitForProcess ph@(ProcessHandle _ delegating_ctlc _) = lockWaitpid $ do
             OpenExtHandle ph' job' -> do
               closePHANDLE ph'
               closePHANDLE job'
-              when delegating_ctlc $
-                endDelegateControlC e
               return (ClosedHandle e, e)
+        -- omit endDelegateControlC since it's a no-op on Windows
         return e'
 #else
     OpenExtHandle _ _job ->
@@ -730,7 +771,8 @@ still running, 'Nothing' is returned.  If the process has exited, then
 @'Just' e@ is returned where @e@ is the exit code of the process.
 
 On Unix systems, see 'waitForProcess' for the meaning of exit codes
-when the process died as the result of a signal.
+when the process died as the result of a signal. May throw
+'UserInterrupt' when using 'delegate_ctlc'.
 -}
 
 getProcessExitCode :: ProcessHandle -> IO (Maybe ExitCode)
@@ -753,6 +795,8 @@ getProcessExitCode ph@(ProcessHandle _ delegating_ctlc _) = tryLockWaitpid $ do
                         let e  | code == 0 = ExitSuccess
                                | otherwise = ExitFailure (fromIntegral code)
                         return (ClosedHandle e, (Just e, True))
+  -- endDelegateControlC after closing the handle, since it
+  -- may throw UserInterrupt
   case m_e of
     Just e | was_open && delegating_ctlc -> endDelegateControlC e
     _                                    -> return ()
@@ -806,7 +850,7 @@ terminateProcess ph = do
     case p_ of
       ClosedHandle  _ -> return ()
 #if defined(WINDOWS)
-      OpenExtHandle{} -> terminateJob ph 1 >> return ()
+      OpenExtHandle{} -> terminateJobUnsafe p_ 1 >> return ()
 #else
       OpenExtHandle{} -> error "terminateProcess with OpenExtHandle should not happen on POSIX."
 #endif
@@ -820,6 +864,19 @@ terminateProcess ph = do
 -- ----------------------------------------------------------------------------
 -- Interface to C bits
 
+#if defined(wasm32_HOST_ARCH)
+
+c_terminateProcess :: PHANDLE -> IO CInt
+c_terminateProcess _ = ioError (ioeSetLocation unsupportedOperation "terminateProcess")
+
+c_getProcessExitCode :: PHANDLE -> Ptr CInt -> IO CInt
+c_getProcessExitCode _ _ = ioError (ioeSetLocation unsupportedOperation "getProcessExitCode")
+
+c_waitForProcess :: PHANDLE -> Ptr CInt -> IO CInt
+c_waitForProcess _ _ = ioError (ioeSetLocation unsupportedOperation "waitForProcess")
+
+#else
+
 foreign import ccall unsafe "terminateProcess"
   c_terminateProcess
         :: PHANDLE
@@ -837,6 +894,7 @@ foreign import ccall interruptible "waitForProcess" -- NB. safe - can block
         -> Ptr CInt
         -> IO CInt
 
+#endif
 
 -- ----------------------------------------------------------------------------
 -- Old deprecated variants
