@@ -88,7 +88,7 @@ data CreateProcess = CreateProcess{
   std_in       :: StdStream,               -- ^ How to determine stdin
   std_out      :: StdStream,               -- ^ How to determine stdout
   std_err      :: StdStream,               -- ^ How to determine stderr
-  close_fds    :: Bool,                    -- ^ Close all file descriptors except stdin, stdout and stderr in the new process (on Windows, only works if std_in, std_out, and std_err are all Inherit). This implementation will call close an every fd from 3 to the maximum of open files, which can be slow for high maximum of open files.
+  close_fds    :: Bool,                    -- ^ Close all file descriptors except stdin, stdout and stderr in the new process (on Windows, only works if std_in, std_out, and std_err are all Inherit). This implementation will call close on every fd from 3 to the maximum of open files, which can be slow for high maximum of open files.
   create_group :: Bool,                    -- ^ Create a new process group
   delegate_ctlc:: Bool,                    -- ^ Delegate control-C handling. Use this for interactive console processes to let them handle control-C themselves (see below for details).
                                            --
@@ -190,6 +190,9 @@ data StdStream
 -- ProcessHandle type
 
 data ProcessHandle__ = OpenHandle { phdlProcessHandle :: PHANDLE }
+                     -- | 'OpenExtHandle' is only applicable for
+                     -- Windows platform. It represents [Job
+                     -- Objects](https://learn.microsoft.com/en-us/windows/win32/procthread/job-objects).
                      | OpenExtHandle { phdlProcessHandle :: PHANDLE
                                      -- ^ the process
                                      , phdlJobHandle     :: PHANDLE
@@ -292,6 +295,7 @@ mbPipeHANDLE CreatePipe pfd  mode =
   do raw_handle <- peek pfd
      let hwnd  = fromHANDLE raw_handle :: Io NativeHandle
          ident = "hwnd:" ++ show raw_handle
-     Just <$> mkHandleFromHANDLE hwnd Stream ident mode Nothing
+     enc <- fmap Just getLocaleEncoding
+     Just <$> mkHandleFromHANDLE hwnd Stream ident mode enc
 mbPipeHANDLE _std      _pfd _mode = return Nothing
 #endif
