@@ -14,20 +14,16 @@ main = do
 
   -- shell kills itself with SIGINT,
   -- delegation off, exit code (death by signal) reported as normal
-  do let script = intercalate "; "
-                    [ "kill -INT $$"
-                    , "exit 42" ]
-     (_,_,_,p) <- createProcess (shell script) { delegate_ctlc = False }
+  do let script = "./process011_c"
+     (_,_,_,p) <- createProcess (proc script []) { delegate_ctlc = False }
      waitForProcess p >>= print
 
   putStrLn "===================== test 2"
 
   -- shell kills itself with SIGINT,
   -- delegation on, so expect to throw UserInterrupt
-  do let script = intercalate "; "
-                    [ "kill -INT $$"
-                    , "exit 42" ]
-     (_,_,_,p) <- createProcess (shell script) { delegate_ctlc = True }
+  do let script = "./process011_c"
+     (_,_,_,p) <- createProcess (proc script []) { delegate_ctlc = True }
      (waitForProcess p >>= print)
        `catchUserInterrupt` \e -> putStrLn $ "caught: " ++ show e
 
