@@ -65,11 +65,23 @@ void releaseAllCapabilities(uint32_t n, Capability *keep_cap, Task *task);
 /* The state of the scheduler.  This is used to control the sequence
  * of events during shutdown.  See Note [shutdown] in Schedule.c.
  */
-#define SCHED_RUNNING       0  /* running as normal */
-#define SCHED_INTERRUPTING  1  /* before threads are deleted */
-#define SCHED_SHUTTING_DOWN 2  /* final shutdown */
+enum SchedState {
+    SCHED_RUNNING       = 0,  /* running as normal */
+    SCHED_INTERRUPTING  = 1,  /* before threads are deleted */
+    SCHED_SHUTTING_DOWN = 2,  /* final shutdown */
+};
 
-extern volatile StgWord sched_state;
+extern StgWord sched_state;
+
+INLINE_HEADER void setSchedState(enum SchedState ss)
+{
+    SEQ_CST_STORE_ALWAYS(&sched_state, (StgWord) ss);
+}
+
+INLINE_HEADER enum SchedState getSchedState(void)
+{
+    return (enum SchedState) SEQ_CST_LOAD_ALWAYS(&sched_state);
+}
 
 /*
  * flag that tracks whether we have done any execution in this time
@@ -87,23 +99,41 @@ extern volatile StgWord sched_state;
  * If the scheduler finds ACTIVITY_DONE_GC and it has a thread to run,
  * it enables the timer again with startTimer().
  */
-#define ACTIVITY_YES      0
-  // the RTS is active
-#define ACTIVITY_MAYBE_NO 1
-  // no activity since the last timer signal
-#define ACTIVITY_INACTIVE 2
-  // RtsFlags.GcFlags.idleGCDelayTime has passed with no activity
-#define ACTIVITY_DONE_GC  3
-  // like ACTIVITY_INACTIVE, but we've done a GC too (if idle GC is
-  // enabled) and the interval timer is now turned off.
+enum RecentActivity {
+    // the RTS is active
+    ACTIVITY_YES      = 0,
+    // no activity since the last timer signal
+    ACTIVITY_MAYBE_NO = 1,
+    // RtsFlags.GcFlags.idleGCDelayTime has passed with no activity
+    ACTIVITY_INACTIVE = 2,
+    // like ACTIVITY_INACTIVE, but we've done a GC too (if idle GC is
+    // enabled) and the interval timer is now turned off.
+    ACTIVITY_DONE_GC  = 3,
+};
 
 /* Recent activity flag.
  * Locks required  : Transition from MAYBE_NO to INACTIVE
  * happens in the timer signal, so it is atomic.  Transition from
  * INACTIVE to DONE_GC happens under sched_mutex.  No lock required
  * to set it to ACTIVITY_YES.
+ *
+ * N.B. we must always use atomics here since even in the non-threaded runtime
+ * the timer may be provided via a signal.
  */
-extern volatile StgWord recent_activity;
+extern StgWord recent_activity;
+
+INLINE_HEADER enum RecentActivity
+setRecentActivity(enum RecentActivity new_value)
+{
+    StgWord old = SEQ_CST_XCHG_ALWAYS((StgPtr) &recent_activity, (StgWord) new_value);
+    return (enum RecentActivity) old;
+}
+
+INLINE_HEADER enum RecentActivity
+getRecentActivity(void)
+{
+    return RELAXED_LOAD_ALWAYS(&recent_activity);
+}
 
 /* Thread queues.
  * Locks required  : sched_mutex
@@ -130,7 +160,7 @@ void resurrectThreads (StgTSO *);
 
 #if !IN_STG_CODE
 
-/* END_TSO_QUEUE and friends now defined in includes/stg/MiscClosures.h */
+/* END_TSO_QUEUE and friends now defined in rts/include/stg/MiscClosures.h */
 
 /* Add a thread to the end of the run queue.
  * NOTE: tso->link should be END_TSO_QUEUE before calling this macro.
