@@ -11,12 +11,10 @@
 
 #include "BeginPrivate.h"
 
-#if defined(DEBUG)
 void    initThreadLabelTable (void);
 void    freeThreadLabelTable (void);
-void *  lookupThreadLabel    (StgWord key);
-void    removeThreadLabel    (StgWord key);
-#endif
+void *  lookupThreadLabel    (StgThreadID key);
+void    removeThreadLabel    (StgThreadID key);
 void    labelThread          (Capability *cap,
                               StgTSO     *tso,
                               char       *label);
