@@ -6,7 +6,7 @@
  *
  * --------------------------------------------------------------------------*/
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "Capability.h"
@@ -176,7 +176,7 @@ cmp_thread(StgPtr tso1, StgPtr tso2)
  *
  * This is used in the implementation of Show for ThreadIds.
  * ------------------------------------------------------------------------ */
-long
+StgThreadID
 rts_getThreadId(StgPtr tso)
 {
   return ((StgTSO *)tso)->id;
@@ -278,15 +278,14 @@ tryWakeupThread (Capability *cap, StgTSO *tso)
         msg->tso = tso;
         SET_HDR(msg, &stg_MSG_TRY_WAKEUP_info, CCS_SYSTEM);
         sendMessage(cap, tso->cap, (Message*)msg);
-        debugTraceCap(DEBUG_sched, cap, "message: try wakeup thread %ld on cap %d",
-                      (W_)tso->id, tso->cap->no);
+        debugTraceCap(DEBUG_sched, cap, "message: try wakeup thread %"
+                      FMT_StgThreadID " on cap %d", tso->id, tso->cap->no);
         return;
     }
 #endif
 
     switch (tso->why_blocked)
     {
-    case BlockedOnIOCompletion:
     case BlockedOnMVar:
     case BlockedOnMVarRead:
     {
@@ -305,8 +304,9 @@ tryWakeupThread (Capability *cap, StgTSO *tso)
         i = lockClosure(tso->block_info.closure);
         unlockClosure(tso->block_info.closure, i);
         if (i != &stg_MSG_NULL_info) {
-            debugTraceCap(DEBUG_sched, cap, "thread %ld still blocked on throwto (%p)",
-                          (W_)tso->id, tso->block_info.throwto->header.info);
+            debugTraceCap(DEBUG_sched, cap, "thread %" FMT_StgThreadID " still "
+                          "blocked on throwto (%p)", tso->id,
+                          tso->block_info.throwto->header.info);
             return;
         }
 
@@ -409,9 +409,8 @@ checkBlockingQueues (Capability *cap, StgTSO *tso)
     StgBlockingQueue *bq, *next;
     StgClosure *p;
 
-    debugTraceCap(DEBUG_sched, cap,
-                  "collision occurred; checking blocking queues for thread %ld",
-                  (W_)tso->id);
+    debugTraceCap(DEBUG_sched, cap, "collision occurred; checking blocking "
+                  "queues for thread %" FMT_StgThreadID, tso->id);
 
     for (bq = tso->bq; bq != (StgBlockingQueue*)END_TSO_QUEUE; bq = next) {
         next = bq->link;
@@ -562,16 +561,16 @@ threadStackOverflow (Capability *cap, StgTSO *tso)
         }
 
         debugTrace(DEBUG_gc,
-                   "threadStackOverflow of TSO %ld (%p): stack too large (now %ld; max is %ld)",
-                   (long)tso->id, tso, (long)tso->stackobj->stack_size,
-                   RtsFlags.GcFlags.maxStkSize);
+                   "threadStackOverflow of TSO %" FMT_StgThreadID " (%p): stack"
+                   " too large (now %ld; max is %ld)", tso->id, tso,
+                   (long)tso->stackobj->stack_size, RtsFlags.GcFlags.maxStkSize);
         IF_DEBUG(gc,
                  /* If we're debugging, just print out the top of the stack */
                  printStackChunk(tso->stackobj->sp,
                                  stg_min(tso->stackobj->stack + tso->stackobj->stack_size,
                                          tso->stackobj->sp+64)));
 
-        // Note [Throw to self when masked], also #767 and #8303.
+        // See Note [Throw to self when masked], also #767 and #8303.
         throwToSelf(cap, tso, (StgClosure *)stackOverflow_closure);
         return;
     }
@@ -888,8 +887,6 @@ printThreadBlockage(StgTSO *tso)
   case BlockedOnMVarRead:
     debugBelch("is blocked on atomic MVar read @ %p", tso->block_info.closure);
     break;
-  case BlockedOnIOCompletion:
-    debugBelch("is blocked on I/O Completion port @ %p", tso->block_info.closure);
     break;
   case BlockedOnBlackHole:
       debugBelch("is blocked on a black hole %p",
@@ -914,8 +911,8 @@ printThreadBlockage(StgTSO *tso)
     debugBelch("is blocked on an STM operation");
     break;
   default:
-    barf("printThreadBlockage: strange tso->why_blocked: %d for TSO %ld (%p)",
-         tso->why_blocked, (long)tso->id, tso);
+    barf("printThreadBlockage: strange tso->why_blocked: %d for TSO %"
+         FMT_StgThreadID " (%p)", tso->why_blocked, tso->id, tso);
   }
 }
 
@@ -953,8 +950,8 @@ printAllThreads(void)
 
   debugBelch("all threads:\n");
 
-  for (i = 0; i < n_capabilities; i++) {
-      cap = capabilities[i];
+  for (i = 0; i < getNumCapabilities(); i++) {
+      cap = getCapability(i);
       debugBelch("threads on capability %d:\n", cap->no);
       for (t = cap->run_queue_hd; t != END_TSO_QUEUE; t = t->_link) {
           printThreadStatus(t);
