@@ -7,7 +7,7 @@
  * ---------------------------------------------------------------------------*/
 
 // Not POSIX, due to use of ru_majflt in getPageFaults()
-// #include "PosixSource.h"
+// #include "rts/PosixSource.h"
 
 #include "Rts.h"
 #include "GetTime.h"
@@ -32,6 +32,10 @@
 #include <mach/mach_port.h>
 #endif
 
+#if defined(solaris2_HOST_OS)
+#include <sys/time.h>
+#endif
+
 #if defined(HAVE_GETTIMEOFDAY) && defined(HAVE_GETRUSAGE)
 // we'll implement getProcessCPUTime() and getProcessElapsedTime()
 // separately, using getrusage() and gettimeofday() respectively
@@ -41,7 +45,7 @@ static uint64_t timer_scaling_factor_numer = 0;
 static uint64_t timer_scaling_factor_denom = 0;
 #endif
 
-void initializeTimer()
+void initializeTimer(void)
 {
 #if defined(darwin_HOST_OS)
     mach_timebase_info_data_t info;
@@ -81,6 +85,12 @@ Time getCurrentThreadCPUTime(void)
         sysErrorBelch("getThreadCPUTime");
         stg_exit(EXIT_FAILURE);
     }
+#elif defined(solaris2_HOST_OS)
+    // On OpenSolaris derivatives, the constant CLOCK_THREAD_CPUTIME_ID is
+    // defined in a system header but it isn't actually usable. clock_gettime(2)
+    // always returns EINVAL. Use solaris-specific gethrvtime(3) as an
+    // alternative.
+    return NSToTime(gethrvtime());
 #elif defined(HAVE_CLOCK_GETTIME)        &&  \
        defined(CLOCK_PROCESS_CPUTIME_ID) &&  \
        defined(HAVE_SYSCONF)
