@@ -6,20 +6,20 @@
  *
  * ---------------------------------------------------------------------------*/
 
-#ifndef BLOCK_ALLOC_H
-#define BLOCK_ALLOC_H
+#pragma once
 
 #include "BeginPrivate.h"
 
 bdescr *allocLargeChunk (W_ min, W_ max);
+bdescr *allocLargeChunkOnNode (uint32_t node, W_ min, W_ max);
 
 /* Debugging  -------------------------------------------------------------- */
 
 extern W_ countBlocks       (bdescr *bd);
 extern W_ countAllocdBlocks (bdescr *bd);
-extern void returnMemoryToOS(nat n);
+extern uint32_t returnMemoryToOS(uint32_t n);
 
-#ifdef DEBUG
+#if defined(DEBUG)
 void checkFreeListSanity(void);
 W_   countFreeList(void);
 void markBlocks (bdescr *bd);
@@ -30,5 +30,3 @@ extern W_ n_alloc_blocks;   // currently allocated blocks
 extern W_ hw_alloc_blocks;  // high-water allocated blocks
 
 #include "EndPrivate.h"
-
-#endif /* BLOCK_ALLOC_H */
