@@ -11,7 +11,7 @@
  *
  * ---------------------------------------------------------------------------*/
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "GCThread.h"
@@ -461,7 +461,6 @@ thread_TSO (StgTSO *tso)
         || tso->why_blocked == BlockedOnMVarRead
         || tso->why_blocked == BlockedOnBlackHole
         || tso->why_blocked == BlockedOnMsgThrowTo
-        || tso->why_blocked == BlockedOnIOCompletion
         || tso->why_blocked == NotBlocked
         ) {
         thread_(&tso->block_info.closure);
@@ -979,8 +978,8 @@ compact(StgClosure *static_objects,
 
     // mutable lists
     for (W_ g = 1; g < RtsFlags.GcFlags.generations; g++) {
-        for (W_ n = 0; n < n_capabilities; n++) {
-            for (bdescr *bd = capabilities[n]->mut_lists[g];
+        for (W_ n = 0; n < getNumCapabilities(); n++) {
+            for (bdescr *bd = getCapability(n)->mut_lists[g];
                  bd != NULL; bd = bd->link) {
                 for (P_ p = bd->start; p < bd->free; p++) {
                     thread((StgClosure **)p);
@@ -1025,7 +1024,7 @@ compact(StgClosure *static_objects,
         debugTrace(DEBUG_gc, "update_fwd:  %d", g);
 
         update_fwd(gen->blocks);
-        for (W_ n = 0; n < n_capabilities; n++) {
+        for (W_ n = 0; n < getNumCapabilities(); n++) {
             update_fwd(gc_threads[n]->gens[g].todo_bd);
             update_fwd(gc_threads[n]->gens[g].part_list);
         }
