@@ -1,3 +1,4 @@
+{-# LANGUAGE Haskell2010 #-}
 {-# LANGUAGE RankNTypes, DataKinds, TypeFamilies #-}
 module FunArgs where
 
@@ -22,7 +23,7 @@ h :: forall a b c
   -> b -- ^ Second argument
   -> c -- ^ Third argument
   -> forall d. d -- ^ Result
-h = undefined
+h _ _ _ = undefined
 
 
 i :: forall a (b :: ()) d. (d ~ '())
